# Copyright 2019 Beijing Lenovo Software Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_versionedobjects import base as object_base

from cyborg.db import api as dbapi
from cyborg.objects import base
from cyborg.objects import fields as object_fields


LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class ARQ(base.CyborgObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    # 1.1: v2 API and Nova integration
    VERSION = '1.1'

    dbapi = dbapi.get_instance()
    fields = {
        'id': object_fields.IntegerField(nullable=False),
        'uuid': object_fields.UUIDField(nullable=False),
        'state': object_fields.ARQStateField(nullable=False),
        'device_profile_name': object_fields.StringField(nullable=False),
        'device_profile_group_id':
            object_fields.IntegerField(nullable=False),

        # Fields populated by Nova after scheduling for binding
        'hostname': object_fields.StringField(nullable=True),
        'device_rp_uuid': object_fields.StringField(nullable=True),
        'instance_uuid': object_fields.StringField(nullable=True),
        'project_id': object_fields.StringField(nullable=True),

        # Fields populated by Cyborg after binding
        'attach_handle_type': object_fields.StringField(nullable=True),
        'attach_handle_uuid': object_fields.StringField(nullable=True),
        'attach_handle_info': object_fields.DictOfStringsField(nullable=True),
    }

    @staticmethod
    def _from_db_object(arq, db_extarq):
        """Converts an ARQ to a formal object.

        :param arq: An object of the class ARQ
        :param db_extarq: A DB model of the object
        :return: The object of the class with the database entity added
        """
        ahi = db_extarq['attach_handle_info']
        if ahi is not None and ahi != '':
            d = jsonutils.loads(ahi)
            db_extarq['attach_handle_info'] = d
        else:
            db_extarq['attach_handle_info'] = {}

        for field in arq.fields:
            arq[field] = db_extarq.get(field)
        arq.obj_reset_changes()
        return arq



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import traceback


def get_exception_handler(debugger_name):
    debugger = _get_debugger(debugger_name)
    return functools.partial(_exception_handler, debugger)


def _get_debugger(debugger_name):
    try:
        debugger = __import__(debugger_name)
    except ImportError:
        raise ValueError("can't import %s module as a post mortem debugger" %
                         debugger_name)
    if 'post_mortem' in dir(debugger):
        return debugger
    else:
        raise ValueError("%s is not a supported post mortem debugger" %
                         debugger_name)


def _exception_handler(debugger, exc_info):
    """Exception handler enabling post-mortem debugging.
    A class extending testtools.TestCase can add this handler in setUp():
        self.addOnException(post_mortem_debug.exception_handler)
    When an exception occurs, the user will be dropped into a debugger
    session in the execution environment of the failure.
    Frames associated with the testing framework are excluded so that
    the post-mortem session for an assertion failure will start at the
    assertion call (e.g. self.assertTrue) rather than the framework code
    that raises the failure exception (e.g. the assertTrue method).
    """
    tb = exc_info[2]
    ignored_traceback = get_ignored_traceback(tb)
    if ignored_traceback:
        tb = FilteredTraceback(tb, ignored_traceback)
    traceback.print_exception(exc_info[0], exc_info[1], tb)
    debugger.post_mortem(tb)


def get_ignored_traceback(tb):
    """Retrieve the first traceback of an ignored trailing chain.
    Given an initial traceback, find the first traceback of a trailing
    chain of tracebacks that should be ignored.  The criteria for
    whether a traceback should be ignored is whether its frame's
    globals include the __unittest marker variable. This criteria is
    culled from:
        unittest.TestResult._is_relevant_tb_level
    For example:
       tb.tb_next => tb0.tb_next => tb1.tb_next
    - If no tracebacks were to be ignored, None would be returned.
    - If only tb1 was to be ignored, tb1 would be returned.
    - If tb0 and tb1 were to be ignored, tb0 would be returned.
    - If either of only tb or only tb0 was to be ignored, None would
      be returned because neither tb or tb0 would be part of a
      trailing chain of ignored tracebacks.
    """
    # Turn the traceback chain into a list
    tb_list = []
    while tb:
        tb_list.append(tb)
        tb = tb.tb_next

    # Find all members of an ignored trailing chain
    ignored_tracebacks = []
    for tb in reversed(tb_list):
        if '__unittest' in tb.tb_frame.f_globals:
            ignored_tracebacks.append(tb)
        else:
            break

    # Return the first member of the ignored trailing chain
    if ignored_tracebacks:
        return ignored_tracebacks[-1]


class FilteredTraceback(object):
    """Wraps a traceback to filter unwanted frames."""

    def __init__(self, tb, filtered_traceback):
        """Constructor.
        :param tb: The start of the traceback chain to filter.
        :param filtered_traceback: The first traceback of a trailing
               chain that is to be filtered.
        """
        self._tb = tb
        self.tb_lasti = self._tb.tb_lasti
        self.tb_lineno = self._tb.tb_lineno
        self.tb_frame = self._tb.tb_frame
        self._filtered_traceback = filtered_traceback

    @property
    def tb_next(self):
        tb_next = self._tb.tb_next
        if tb_next and tb_next != self._filtered_traceback:
            return FilteredTraceback(tb_next, self._filtered_traceback)



# Copyright 2019 Intel, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg.common import exception
from cyborg import objects
from cyborg.tests.unit.db import base
from cyborg.tests.unit import fake_device


class TestDeviceObject(base.DbTestCase):

    def setUp(self):
        super(TestDeviceObject, self).setUp()
        self.fake_device = fake_device.get_db_devices()[0]

    def test_get(self):
        uuid = self.fake_device['uuid']
        with mock.patch.object(self.dbapi, 'device_get',
                               autospec=True) as mock_device_get:
            mock_device_get.return_value = self.fake_device
            device = objects.Device.get(self.context, uuid)
            mock_device_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(self.context, device._context)

    def test_get_by_id(self):
        device_id = self.fake_device['id']
        with mock.patch.object(self.dbapi, 'device_get_by_id',
                               autospec=True) as mock_device_get_by_id:
            mock_device_get_by_id.return_value = self.fake_device
            device = objects.Device.get_by_device_id(self.context, device_id)
            mock_device_get_by_id.assert_called_once_with(
                self.context, device_id)
            self.assertEqual(self.context, device._context)

    def test_get_by_non_existed_id(self):
        device_id = self.fake_device['id']
        with mock.patch.object(self.dbapi, 'device_get_by_id',
                               autospec=True) as mock_device_get_by_id:
            mock_device_get_by_id.side_effect = exception.ResourceNotFound(
                resource='Device', msg='with uuid=%s' % device_id)
            self.assertRaises(exception.ResourceNotFound,
                              objects.Device.get_by_device_id,
                              self.context,
                              device_id)

    @mock.patch.object(objects.Device, 'list')
    def test_get_by_hostname(self, mock_list):
        hostname = self.fake_device['hostname']
        dev_filter = {'hostname': hostname}
        mock_list.return_value = [self.fake_device]
        devices = objects.Device.get_list_by_hostname(self.context, hostname)
        mock_list.assert_called_once_with(self.context, dev_filter)
        self.assertEqual(hostname, devices[0]['hostname'])

        # test objects.Device.list return [] when hostname is None.
        mock_list.return_value = []
        hostname = None
        dev_filter = {'hostname': hostname}
        devices = objects.Device.get_list_by_hostname(self.context, hostname)
        mock_list.assert_called_with(self.context, dev_filter)
        self.assertEqual([], devices)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'device_list',
                               autospec=True) as mock_device_list:
            mock_device_list.return_value = [self.fake_device]
            devices = objects.Device.list(self.context)
            self.assertEqual(1, mock_device_list.call_count)
            self.assertEqual(1, len(devices))
            self.assertIsInstance(devices[0], objects.Device)
            self.assertEqual(self.context, devices[0]._context)

    def test_list_with_filter(self):
        with mock.patch.object(self.dbapi, 'device_list_by_filters',
                               autospec=True) as mock_device_with_filter_list:
            mock_device_with_filter_list.return_value = [self.fake_device]
            filters = {'limit': 1}
            devices = objects.Device.list(self.context, filters)
            self.assertEqual(1, mock_device_with_filter_list.call_count)
            self.assertEqual(1, len(devices))
            self.assertIsInstance(devices[0], objects.Device)
            mock_device_with_filter_list.assert_called_once_with(
                self.context,
                {},
                sort_dir='desc',
                sort_key='created_at',
                limit=1,
                marker=None,
                )

    def test_create(self):
        with mock.patch.object(self.dbapi, 'device_create',
                               autospec=True) as mock_device_create:
            mock_device_create.return_value = self.fake_device
            device = objects.Device(self.context,
                                    **self.fake_device)
            device.create(self.context)
            mock_device_create.assert_called_once_with(
                self.context, self.fake_device)
            self.assertEqual(self.context, device._context)

    def test_destroy(self):
        uuid = self.fake_device['uuid']
        with mock.patch.object(self.dbapi, 'device_get',
                               autospec=True) as mock_device_get:
            mock_device_get.return_value = self.fake_device
            with mock.patch.object(self.dbapi, 'device_delete',
                                   autospec=True) as mock_device_delete:
                device = objects.Device.get(self.context, uuid)
                device.destroy(self.context)
                mock_device_delete.assert_called_once_with(self.context,
                                                           uuid)
                self.assertEqual(self.context, device._context)

    def test_update(self):
        uuid = self.fake_device['uuid']
        with mock.patch.object(self.dbapi, 'device_get',
                               autospec=True) as mock_device_get:
            mock_device_get.return_value = self.fake_device
            with mock.patch.object(self.dbapi, 'device_update',
                                   autospec=True) as mock_device_update:
                fake = self.fake_device
                fake["vendor_board_info"] = "new_vendor_board_info"
                mock_device_update.return_value = fake
                device = objects.Device.get(self.context, uuid)
                device.vendor_board_info = 'new_vendor_board_info'
                device.save(self.context)
                mock_device_get.assert_called_once_with(self.context,
                                                        uuid)
                mock_device_update.assert_called_once_with(
                    self.context, uuid,
                    {'vendor_board_info': 'new_vendor_board_info'})
                self.assertEqual(self.context, device._context)

    def test_device_type(self):
        for t in ["GPU", "FPGA", "AICHIP"]:
            device = objects.Device(self.context, type=t)
            self.assertEqual(self.context, device._context)
        # Invaild type will raise ValueError
        self.assertRaises(ValueError, objects.Device,
                          self.context, type='OTHER_TYPE')



# Copyright 2021 Inspur, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Cyborg Xilinx FPGA driver implementation.
"""

import re

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_serialization import jsonutils

from cyborg.accelerator.common import utils
from cyborg.common import constants
from cyborg.conf import CONF
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device
from cyborg.privsep import sys_admin_pctxt

LOG = logging.getLogger(__name__)

XILINX_FPGA_FLAGS = ["Xilinx Corporation Device",
                     "Processing accelerators"]

XILINX_FPGA_INFO_PATTERN = re.compile(
    r"(?P<pci_addr>[0-9a-fA-F]{4}:[0-9a-fA-F]{2}:"
    r"[0-9a-fA-F]{2}\.[0-9a-fA-F]) "
    r"(?P<controller>.*) [\[].*]: (?P<model>.*) .*"
    r"[\[](?P<vendor_id>[0-9a-fA-F]"
    r"{4}):(?P<product_id>[0-9a-fA-F]{4})].*")

XILINX_PF_MAPS = {"mgmt": "xclmgmt", "user": "xocl"}

VENDOR_MAPS = {"10ee": "xilinx"}


@sys_admin_pctxt.entrypoint
def lspci_privileged(cmd):
    return processutils.execute(*cmd)


def get_pci_devices(pci_flags, vendor_id=None):
    device_for_vendor_out = []
    all_device_out = []
    cmd = ['lspci', '-nnn', '-D']
    lspci_out = lspci_privileged(cmd)[0].split('\n')
    for i in range(len(lspci_out)):
        if all(x in lspci_out[i] for x in pci_flags):
            all_device_out.append(lspci_out[i])
            if vendor_id and vendor_id in lspci_out[i]:
                device_for_vendor_out.append(lspci_out[i])
    return device_for_vendor_out if vendor_id else all_device_out


def _generate_traits(vendor_id, product_id):
    """Generate traits for FPGAs.
    : param vendor_id: vendor_id of FPGA
    : param product_id: product_id of FPGA
    Example FPGA traits:
    {traits:["CUSTOM_FPGA_XILINX", "CUSTOM_FPGA_PRODUCT_ID_5001"]}
    """
    traits = []
    traits.append("CUSTOM_FPGA_" + VENDOR_MAPS.get(vendor_id, "").upper())
    traits.append("CUSTOM_FPGA_PRODUCT_ID_" + product_id.upper())
    return {"traits": traits}


def _get_pf_type(device):
    cmd = ['lspci', '-k', '-s', device]
    result = lspci_privileged(cmd)[0]
    for k, v in XILINX_PF_MAPS.items():
        if v in result:
            return k


def _combine_device_by_pci_func(pci_devices):
    fpga_devices = []
    for pci_dev in pci_devices:
        m = XILINX_FPGA_INFO_PATTERN.match(pci_dev)
        if m:
            pci_dict = m.groupdict()
            LOG.debug('Xilinx fpga pci device in dict: %s', pci_dict)
            # 0000:3b:00.0/1
            is_existed = False
            new_addr = pci_dict.get('pci_addr')
            for fpga in fpga_devices:
                existed_addr = fpga.get('pci_addr')[0]
                # compare domain:bus:slot
                if existed_addr and \
                        new_addr.split('.')[0] == existed_addr.split('.')[0]:
                    fpga.update({'pci_addr': [existed_addr, new_addr]})
                    is_existed = True
            if not is_existed:
                traits = _generate_traits(pci_dict["vendor_id"],
                                          pci_dict["product_id"])
                pci_dict["rc"] = constants.RESOURCES["FPGA"]
                pci_dict.update(traits)
                pci_dict.update({'pci_addr': [new_addr]})
                fpga_devices.append(pci_dict)
    return fpga_devices


def _generate_controlpath_id(fpga):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(fpga['pci_addr'][0])
    return driver_cpid


def _generate_dep_list(fpga):
    driver_dep = driver_deployable.DriverDeployable()
    driver_dep.attribute_list = _generate_attribute_list(fpga)
    driver_dep.attach_handle_list = []
    driver_dep.name = CONF.host + '_' + fpga["pci_addr"][0]
    driver_dep.driver_name = VENDOR_MAPS.get(fpga["vendor_id"], '').upper()
    driver_dep.num_accelerators = 1
    driver_dep.attach_handle_list = _generate_attach_handle(fpga)
    return [driver_dep]


def _generate_attribute_list(fpga):
    attr_list = []
    for k, v in fpga.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key, driver_attr.value = k, v
            attr_list.append(driver_attr)
        if k == "traits":
            values = fpga.get(k, [])
            for index, val in enumerate(values):
                driver_attr = driver_attribute.DriverAttribute(
                    key="trait" + str(index), value=val)
                attr_list.append(driver_attr)
    return attr_list


def _generate_attach_handle(fpga):
    driver_ahs = []
    for addr in fpga.get('pci_addr'):
        driver_ah = driver_attach_handle.DriverAttachHandle()
        driver_ah.attach_type = constants.AH_TYPE_PCI
        driver_ah.attach_info = utils.pci_str_to_json(addr)
        driver_ah.in_use = False
        driver_ahs.append(driver_ah)
    return driver_ahs


def fpga_tree():
    fpga_list = []
    fpga_pci_devices = get_pci_devices(XILINX_FPGA_FLAGS)
    LOG.debug("Xilinx fpga devices from lspci: %s", fpga_pci_devices)
    # In the return pci devices, mgmt pf and user pf are two entries.
    # Now only when binding both to vm, end user can program it.
    # So combine these two entries into one device.
    for fpga in _combine_device_by_pci_func(fpga_pci_devices):
        driver_device_obj = driver_device.DriverDevice()
        driver_device_obj.vendor = fpga["vendor_id"]
        driver_device_obj.model = fpga.get('model', 'miss model info')
        std_board_info = {'product_id': fpga.get('product_id'),
                          'controller': fpga.get('controller')}
        vendor_board_info = {
            'vendor_info': fpga.get('vendor_info', 'fpga_vb_info')}
        driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
        driver_device_obj.vendor_board_info = \
            jsonutils.dumps(vendor_board_info)
        driver_device_obj.type = constants.DEVICE_FPGA
        driver_device_obj.stub = fpga.get('stub', False)
        driver_device_obj.controlpath_id = _generate_controlpath_id(fpga)
        driver_device_obj.deployable_list = _generate_dep_list(fpga)
        fpga_list.append(driver_device_obj)
    return fpga_list



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""legacy old_policies, the following old_policies will be removed once
   new policies are implemented.
"""
from oslo_policy import policy
# NOTE: to follow policy-in-code spec, we define defaults for
#       the granular policies in code, rather than in policy.yaml.
#       All of these may be overridden by configuration, but we can
#       depend on their existence throughout the code.

accelerator_request_policies = [
    policy.RuleDefault('cyborg:arq:get_all',
                       'rule:default',
                       description='Retrieve accelerator request records.'),
    policy.RuleDefault('cyborg:arq:get_one',
                       'rule:default',
                       description='Get an accelerator request record.'),
    policy.RuleDefault('cyborg:arq:create',
                       'rule:allow',
                       description='Create accelerator request records.'),
    policy.RuleDefault('cyborg:arq:delete',
                       'rule:default',
                       description='Delete accelerator request records.'),
    policy.RuleDefault('cyborg:arq:update',
                       'rule:default',
                       description='Update accelerator request records.'),
]

device_policies = [
    policy.RuleDefault('cyborg:device:get_one',
                       'rule:allow',
                       description='Show device detail'),
    policy.RuleDefault('cyborg:device:get_all',
                       'rule:allow',
                       description='Retrieve all device records'),
    policy.RuleDefault('cyborg:device:disable',
                       'rule:admin_api',
                       description='Disable a device'),
    policy.RuleDefault('cyborg:device:enable',
                       'rule:admin_api',
                       description='Enable a device'),
]

deployable_policies = [
    policy.RuleDefault('cyborg:deployable:get_one',
                       'rule:allow',
                       description='Show deployable detail'),
    policy.RuleDefault('cyborg:deployable:get_all',
                       'rule:allow',
                       description='Retrieve all deployable records'),
    policy.RuleDefault('cyborg:deployable:program',
                       'rule:allow',
                       description='FPGA programming.'),
]

attribute_policies = [
    policy.RuleDefault('cyborg:attribute:get_one',
                       'rule:allow',
                       description='Show attribute detail'),
    policy.RuleDefault('cyborg:attribute:get_all',
                       'rule:allow',
                       description='Retrieve all attribute records'),
    policy.RuleDefault('cyborg:attribute:create',
                       'rule:allow',
                       description='Create an attribute record'),
    policy.RuleDefault('cyborg:attribute:delete',
                       'rule:allow',
                       description='Delete attribute records.'),
]

fpga_policies = [
    policy.RuleDefault('cyborg:fpga:get_one',
                       'rule:allow',
                       description='Show fpga detail'),
    policy.RuleDefault('cyborg:fpga:get_all',
                       'rule:allow',
                       description='Retrieve all fpga records'),
    policy.RuleDefault('cyborg:fpga:update',
                       'rule:allow',
                       description='Update fpga records'),
]



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from cyborg.common.i18n import _


opts = [
    cfg.ListOpt('enabled_drivers',
                default=['fake_driver'],
                help=_('The accelerator drivers enabled on this agent. Such '
                       'as intel_fpga_driver, inspur_fpga_driver,'
                       'nvidia_gpu_driver, intel_qat_driver,'
                       'inspur_nvme_ssd_driver, xilinx_fpga_driver, etc.')),
]

opt_group = cfg.OptGroup(name='agent',
                         title='Options for the cyborg-agent service')


AGENT_OPTS = (opts)


def register_opts(conf):
    conf.register_group(opt_group)
    conf.register_opts(opts, group=opt_group)


def list_opts():
    return {
        opt_group: AGENT_OPTS
    }



# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Utils for FPGA driver.
"""

import glob
import re


VENDORS = ["intel", "inspur"]  # can extend, such as ["intel", "xilinx"]

SYS_FPGA_PATH = "/sys/class/fpga"
VENDORS_PATTERN = re.compile("|".join(["(%s)" % v for v in VENDORS]))


def discover_vendors():
    vendors = set()
    for p in glob.glob1(SYS_FPGA_PATH, "*"):
        m = VENDORS_PATTERN.match(p)
        if m:
            vendors.add(m.group())
    return list(vendors)



# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Unit tests for the DB api."""

import datetime

from oslo_utils import timeutils

from cyborg.db import api as dbapi
from cyborg.db.sqlalchemy import api as sqlalchemyapi
from cyborg.tests.unit.db import base


def _quota_reserve(context, project_id):
    """Create sample QuotaUsage and Reservation objects.

    There is no method db.quota_usage_create(), so we have to use
    db.quota_reserve() for creating QuotaUsage objects.

    Returns reservations uuids.

    """
    sqlalchemy_api = sqlalchemyapi.get_backend()
    resources = {}
    deltas = {}
    for i, resource in enumerate(('fpga', 'gpu')):
        deltas[resource] = i + 1
    return sqlalchemy_api.quota_reserve(
        context, resources, deltas,
        timeutils.utcnow(), timeutils.utcnow(),
        datetime.timedelta(days=1), project_id
    )


class DBAPIQuotaUsageTestCase(base.DbTestCase):

    """Tests for db.api.quota_usage_* methods."""

    def _test_quota_reserve(self):
        sqlalchemy_api = sqlalchemyapi.get_backend()
        reservations = _quota_reserve(self.context, 'project1')
        self.assertEqual(2, len(reservations))
        quota_usages = sqlalchemy_api._get_quota_usages(self.context,
                                                        'project1')
        result = {'project_id': "project1"}
        for k, v in quota_usages.items():
            result[v.resource] = dict(in_use=v.in_use, reserved=v.reserved)

        self.assertEqual({'project_id': 'project1',
                          'gpu': {'reserved': 2, 'in_use': 0},
                          'fpga': {'reserved': 1, 'in_use': 0}},
                         result)

    def _test__get_quota_usages(self):
        _quota_reserve(self.context, 'project1')
        sqlalchemy_api = sqlalchemyapi.get_backend()
        quota_usages = sqlalchemy_api._get_quota_usages(self.context,
                                                        'project1')

        self.assertEqual(['fpga', 'gpu'],
                         sorted(quota_usages.keys()))

    def _test__get_quota_usages_with_resources(self):
        _quota_reserve(self.context, 'project1')
        sqlalchemy_api = sqlalchemyapi.get_backend()
        quota_usage = sqlalchemy_api._get_quota_usages(
            self.context, 'project1', resources=['gpu'])

        self.assertEqual(['gpu'], list(quota_usage.keys()))


class DBAPIReservationTestCase(base.DbTestCase):

    """Tests for db.api.reservation_* methods."""

    def setUp(self):
        super(DBAPIReservationTestCase, self).setUp()
        self.values = {
            'uuid': 'sample-uuid',
            'project_id': 'project1',
            'resource': 'resource',
            'delta': 42,
            'expire': (timeutils.utcnow() +
                       datetime.timedelta(days=1)),
            'usage': {'id': 1}
        }

    def _test__get_reservation_resources(self):
        sqlalchemy_api = sqlalchemyapi.get_backend()
        reservations = _quota_reserve(self.context, 'project1')
        expected = ['fpga', 'gpu']
        resources = sqlalchemy_api._get_reservation_resources(
            self.context, reservations)
        self.assertEqual(expected, sorted(resources))

    def _test_reservation_commit(self):
        db_api = dbapi.get_instance()
        reservations = _quota_reserve(self.context, 'project1')
        expected = {'project_id': 'project1',
                    'fpga': {'reserved': 1, 'in_use': 0},
                    'gpu': {'reserved': 2, 'in_use': 0},
                    }
        quota_usages = db_api._get_quota_usages(self.context, 'project1')
        result = {'project_id': "project1"}
        for k, v in quota_usages.items():
            result[v.resource] = dict(in_use=v.in_use, reserved=v.reserved)

        self.assertEqual(expected, result)

        db_api.reservation_commit(self.context, reservations, 'project1')
        expected = {'project_id': 'project1',
                    'fpga': {'reserved': 0, 'in_use': 1},
                    'gpu': {'reserved': 0, 'in_use': 2},
                    }
        quota_usages = db_api._get_quota_usages(self.context, 'project1')
        result = {'project_id': "project1"}
        for k, v in quota_usages.items():
            result[v.resource] = dict(in_use=v.in_use,
                                      reserved=v.reserved)
        self.assertEqual(expected, result)



# Copyright 2017 Lenovo, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Accelerator base exception handling. """

import collections
from http import HTTPStatus
from oslo_log import log as logging
from oslo_serialization import jsonutils

from cyborg.common.i18n import _


LOG = logging.getLogger(__name__)


def _ensure_exception_kwargs_serializable(exc_class_name, kwargs):
    """Ensure that kwargs are serializable

    Ensure that all kwargs passed to exception constructor can be passed over
    RPC, by trying to convert them to JSON, or, as a last resort, to string.
    If it is not possible, unserializable kwargs will be removed, letting the
    receiver to handle the exception string as it is configured to.

    :param exc_class_name: an AcceleratorException class name.
    :param kwargs: a dictionary of keyword arguments passed to the exception
        constructor.
    :returns: a dictionary of serializable keyword arguments.
    """
    serializers = [(jsonutils.dumps, _('when converting to JSON')),
                   (str, _('when converting to string'))]
    exceptions = collections.defaultdict(list)
    serializable_kwargs = {}
    for k, v in kwargs.items():
        for serializer, msg in serializers:
            try:
                serializable_kwargs[k] = serializer(v)
                exceptions.pop(k, None)
                break
            except Exception as e:
                exceptions[k].append(
                    '(%(serializer_type)s) %(e_type)s: %(e_contents)s' %
                    {'serializer_type': msg, 'e_contents': e,
                     'e_type': e.__class__.__name__})
    if exceptions:
        LOG.error("One or more arguments passed to the %(exc_class)s "
                  "constructor as kwargs can not be serialized. The "
                  "serialized arguments: %(serialized)s. These "
                  "unserialized kwargs were dropped because of the "
                  "exceptions encountered during their "
                  "serialization:\n%(errors)s",
                  dict(errors=';\n'.join("%s: %s" % (k, '; '.join(v))
                                         for k, v in exceptions.items()),
                       exc_class=exc_class_name,
                       serialized=serializable_kwargs))
        # We might be able to actually put the following keys' values into
        # format string, but there is no guarantee, drop it just in case.
        for k in exceptions:
            del kwargs[k]
    return serializable_kwargs


class AcceleratorException(Exception):
    """Base Accelerator Exception

    To correctly use this class, inherit from it and define
    a '_msg_fmt' property. That message will get printf'd
    with the keyword arguments provided to the constructor.

    If you need to access the message from an exception you should use
    str(exc).
    """
    _msg_fmt = _("An unknown exception occurred.")
    code = HTTPStatus.INTERNAL_SERVER_ERROR
    headers = {}
    safe = False

    def __init__(self, message=None, **kwargs):

        self.kwargs = _ensure_exception_kwargs_serializable(
            self.__class__.__name__, kwargs)

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            if kwargs:
                message = self._msg_fmt % kwargs
            else:
                message = self._msg_fmt

        super(AcceleratorException, self).__init__(message)


class Invalid(AcceleratorException):
    _msg_fmt = _("Unacceptable parameters.")


class InvalidParameterValue(Invalid):
    _msg_fmt = "%(err)s"


class MissingParameterValue(InvalidParameterValue):
    _msg_fmt = "%(err)s"


class InvalidAccelerator(InvalidParameterValue):
    _msg_fmt = "%(err)s"



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Policy Authorize Engine For Cyborg."""
import functools
import sys

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log
from oslo_policy import policy
from oslo_versionedobjects import base as object_base
import pecan
import wsme

from cyborg.common import exception
from cyborg import policies


_ENFORCER = None
CONF = cfg.CONF
LOG = log.getLogger(__name__)


@lockutils.synchronized('policy_enforcer', 'cyborg-')
def init_enforcer(policy_file=None, rules=None,
                  default_rule=None, use_conf=True,
                  suppress_deprecation_warnings=False):
    """Synchronously initializes the policy enforcer
    :param policy_file: Custom policy file to use, if none is specified,
                        `CONF.oslo_policy.policy_file` will be used.
    :param rules: Default dictionary / Rules to use. It will be
                  considered just in the first instantiation.
    :param default_rule: Default rule to use,
                         CONF.oslo_policy.policy_default_rule will
                         be used if none is specified.
    :param use_conf: Whether to load rules from config file.
    :param suppress_deprecation_warnings: Whether to suppress the
                                          deprecation warnings.
    """
    global _ENFORCER

    if _ENFORCER:
        return

    # NOTE: Register defaults for policy-in-code here so that they are
    # loaded exactly once - when this module-global is initialized.
    # Defining these in the relevant API modules won't work
    # because API classes lack singletons and don't use globals.
    _ENFORCER = policy.Enforcer(
        CONF,
        policy_file=policy_file,
        rules=rules,
        default_rule=default_rule,
        use_conf=use_conf)
    if suppress_deprecation_warnings:
        _ENFORCER.suppress_deprecation_warnings = True
    _ENFORCER.register_defaults(policies.list_policies())


def get_enforcer():
    """Provides access to the single accelerator of policy enforcer."""
    global _ENFORCER

    if not _ENFORCER:
        init_enforcer()

    return _ENFORCER


# NOTE: We can't call these methods from within decorators because the
# 'target' and 'creds' parameter must be fetched from the call time
# context-local pecan.request magic variable, but decorators are compiled
# at module-load time.


def authorize(rule, target, creds, do_raise=False, *args, **kwargs):
    """A shortcut for policy.Enforcer.authorize()
    Checks authorization of a rule against the target and credentials, and
    raises an exception if the rule is not defined.
    """
    enforcer = get_enforcer()
    try:
        return enforcer.authorize(rule, target, creds, do_raise=do_raise,
                                  *args, **kwargs)
    except policy.PolicyNotAuthorized:
        raise exception.HTTPForbidden(resource=rule)


# This decorator MUST appear first (the outermost decorator)
# on an API method for it to work correctly
def authorize_wsgi(api_name, act=None, need_target=True):
    """This is a decorator to simplify wsgi action policy rule check.
    :param api_name: The collection name to be evaluate.
    :param act: The function name of wsgi action.
    :param need_target: Whether need target for authorization. Such as,
                        when create some resource , maybe target is not needed.
    example:
        from cyborg.common import authorize_wsgi
        class AcceleratorController(rest.RestController):
            ....
            @authorize_wsgi.authorize_wsgi("cyborg:accelerator",
                                           "create", False)
            @wsme_pecan.wsexpose(Accelerator, body=types.jsontype,
                                 status_code=HTTPStatus.CREATED)
            def post(self, values):
                ...
    """
    def wraper(fn):
        action = '%s:%s' % (api_name, act or fn.__name__)

        # In this authorize method, we return a dict data when authorization
        # fails or exception comes out. Maybe we can consider to use
        # wsme.api.Response in future.
        def return_error(resp_status):
            exception_info = sys.exc_info()
            orig_exception = exception_info[1]
            orig_code = getattr(orig_exception, 'code', None)
            pecan.response.status = orig_code or resp_status
            data = wsme.api.format_exception(
                exception_info,
                pecan.conf.get('wsme', {}).get('debug', False)
            )
            del exception_info
            return data

        @functools.wraps(fn)
        def handle(self, *args, **kwargs):
            context = pecan.request.context
            credentials = context.to_policy_values()
            credentials['is_admin'] = context.is_admin
            if context.system_scope == 'all':
                credentials['system'] = True
            target = {}
            # maybe we can pass "_get_resource" to authorize_wsgi
            if need_target and hasattr(self, "_get_resource"):
                try:
                    resource = getattr(self, "_get_resource")(*args, **kwargs)
                    # just support object, other type will just keep target as
                    # empty, then follow authorize method will fail and throw
                    # an exception
                    if isinstance(resource,
                                  object_base.VersionedObjectDictCompat):
                        target = {'project_id': resource.project_id,
                                  'user_id': resource.user_id}
                except Exception:
                    return return_error(500)
            elif need_target:
                # if developer do not set _get_resource, just set target as
                # empty, then follow authorize method will fail and throw an
                # exception
                target = {}
            else:
                # for create method, before resource exsites, we can check the
                # the credentials with itself.
                target = {'project_id': context.project_id,
                          'user_id': context.user_id}

            try:
                authorize(action, target, credentials, do_raise=True)
            except Exception:
                return return_error(403)

            return fn(self, *args, **kwargs)

        return handle

    return wraper



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg.accelerator.drivers.spdk.nvmf.nvmf import NVMFDRIVER
from cyborg.accelerator.drivers.spdk.util import common_fun
from cyborg.accelerator.drivers.spdk.util.pyspdk.nvmf_client import NvmfTgt
from cyborg.tests import base


class TestNVMFDRIVER(base.TestCase):

    def setUp(self,):
        super(TestNVMFDRIVER, self).setUp()
        self.nvmf_driver = NVMFDRIVER()

    def tearDown(self):
        super(TestNVMFDRIVER, self).tearDown()
        self.vhost_driver = None

    @mock.patch.object(NVMFDRIVER, 'get_one_accelerator')
    def test_discover_accelerator(self, mock_get_one_accelerator):
        expect_accelerator = {
            'server': 'nvmf',
            'bdevs': [{"num_blocks": 131072,
                       "name": "nvme1",
                       "block_size": 512
                       }],
            'subsystems': [{"core": 0,
                            "nqn": "nqn.2018-01.org.nvmexpress.discovery",
                            "hosts": []
                            }]
        }
        alive = mock.Mock(return_value=False)
        self.nvmf_driver.py.is_alive = alive
        check_error = mock.Mock(return_value=False)
        common_fun.check_for_setup_error = check_error
        self.assertFalse(
            mock_get_one_accelerator.called,
            "Failed to discover_accelerator if py not alive."
        )
        alive = mock.Mock(return_value=True)
        self.nvmf_driver.py.is_alive = alive
        check_error = mock.Mock(return_value=True)
        common_fun.check_for_setup_error = check_error
        acce_client = NvmfTgt(self.nvmf_driver.py)
        bdevs_fake = [{"num_blocks": 131072,
                       "name": "nvme1",
                       "block_size": 512
                       }]
        bdev_list = mock.Mock(return_value=bdevs_fake)
        acce_client.get_bdevs = bdev_list
        subsystems_fake = [{"core": 0,
                            "nqn": "nqn.2018-01.org.nvmexpress.discovery",
                            "hosts": []
                            }]
        subsystem_list = mock.Mock(return_value=subsystems_fake)
        acce_client.get_nvmf_subsystems = subsystem_list
        accelerator_fake = {
            'server': self.nvmf_driver.SERVER,
            'bdevs': acce_client.get_bdevs(),
            'subsystems': acce_client.get_nvmf_subsystems()
        }
        success_send = mock.Mock(return_value=accelerator_fake)
        self.nvmf_driver.get_one_accelerator = success_send
        accelerator = self.nvmf_driver.discover_accelerator()
        self.assertEqual(accelerator, expect_accelerator)

    def test_accelerator_list(self):
        expect_accelerators = [{
            'server': 'nvmf',
            'bdevs': [{"num_blocks": 131072,
                       "name": "nvme1",
                       "block_size": 512
                       }],
            'subsystems':
                [{"core": 0,
                  "nqn": "nqn.2018-01.org.nvmexpress.discovery",
                  "hosts": []
                  }]
        },
            {
                'server': 'nvnf_tgt',
                'bdevs': [{"num_blocks": 131072,
                           "name": "nvme1",
                           "block_size": 512
                           }],
                'subsystems':
                    [{"core": 0,
                      "nqn": "nqn.2018-01.org.nvmexpress.discovery",
                      "hosts": []
                      }]
            }
        ]
        success_send = mock.Mock(return_value=expect_accelerators)
        self.nvmf_driver.get_all_accelerators = success_send
        self.assertEqual(self.nvmf_driver.accelerator_list(),
                         expect_accelerators)

    def test_install_accelerator(self):
        pass

    def test_uninstall_accelerator(self):
        pass

    def test_update(self):
        pass

    def test_attach_instance(self):
        pass

    def test_detach_instance(self):
        pass

    def test_delete_subsystem(self):
        pass

    def test_construct_subsystem(self):
        pass



# Copyright 2019 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cyborg.common import exception
from cyborg.common.i18n import _
from cyborg.common import utils
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class NovaAPI(object):
    def __init__(self):
        self.nova_client = utils.get_sdk_adapter('compute')
        self.nova_client.default_microversion = '2.82'

    def _get_acc_changed_events(self, instance_uuid, arq_bind_statuses):
        return [{'name': 'accelerator-request-bound',
                 'server_uuid': instance_uuid,
                 'tag': arq_uuid,
                 'status': arq_bind_status,
                 } for (arq_uuid, arq_bind_status) in arq_bind_statuses]

    def _send_events(self, events):
        """Send events to Nova external events API.

        :param events: List of events to send to Nova.
        :raises: exception.InvalidAPIResponse, on unexpected error
        """
        url = "/os-server-external-events"
        body = {"events": events}
        response = self.nova_client.post(url, json=body)
        # NOTE(Sundar): Response status should always be 200/207. See
        # https://review.opendev.org/#/c/698037/
        if response.status_code == 200:
            LOG.info("Sucessfully sent events to Nova, events: %(events)s",
                     {"events": events})
        elif response.status_code == 207:
            # NOTE(Sundar): If Nova returns per-event code of 422, that
            # is due to a race condition where Nova has not associated
            # the instance with a host yet. See
            # https://bugs.launchpad.net/nova/+bug/1855752
            events = [ev for ev in response.json()['events']]
            event_codes = {ev['code'] for ev in events}
            if len(event_codes) == 1:  # all events have same event code
                if event_codes == {422}:
                    LOG.info('Ignoring Nova notification error that the '
                             'instance %s is not yet associated with a host.',
                             events[0]['server_uuid'])
                else:
                    msg = _('Unexpected event code %(code)s '
                            'for instance %(inst)s')
                    msg = msg % {'code': event_codes.pop(),
                                 'inst': events[0]["server_uuid"]}
                    raise exception.InvalidAPIResponse(
                        service='Nova', api=url[1:], msg=msg)
            else:
                msg = _('All event responses are expected to '
                        'have the same event code. Instance: %(inst)s')
                msg = msg % {'inst': events[0]['server_uuid']}
                raise exception.InvalidAPIResponse(
                    service='Nova', api=url[1:], msg=msg)
        else:
            # Unexpected return code from Nova
            msg = _('Failed to send events %(ev)s: HTTP %(code)s: %(txt)s')
            msg = msg % {'ev': events,
                         'code': response.status_code,
                         'txt': response.text}
            raise exception.InvalidAPIResponse(
                service='Nova', api=url[1:], msg=msg)

    def notify_binding(self, instance_uuid, arq_bind_statuses):
        """Notify Nova that ARQ bindings are resolved for a given instance.

        :param instance_uuid: UUID of the instance whose ARQs are resolved
        :param arq_bind_statuses: List of (arq_uuid, arq_bind_status) tuples
        :returns: None
        """
        events = self._get_acc_changed_events(instance_uuid, arq_bind_statuses)
        self._send_events(events)



# Copyright 2019 Intel Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from cyborg.common import constants
from cyborg.common.constants import ARQ_STATES_TRANSFORM_MATRIX
from cyborg.common import exception
from cyborg.common import nova_client
from cyborg.common import utils
from cyborg.conf import CONF
from cyborg import objects


LOG = logging.getLogger(__name__)


class ExtARQJobMixin(object):
    """Mixin Class for ExtARQ async job management."""

    def _bind_job(self, context, deployable):
        """The bind process of an accelerator."""
        check_extra_job = getattr(self, "_need_extra_bind_job", None)
        need_job = False
        if check_extra_job:
            need_job = check_extra_job(context, deployable)
        if getattr(self.bind, "is_job", False) and need_job:
            LOG.info("Start job for ARQ(%s) bind.", self.arq.uuid)
            works = utils.ThreadWorks()
            job = works.spawn(self.bind, context, deployable)
            return job
        else:
            LOG.info("ARQ(%s) bind process is instant.", self.arq.uuid)
            self.bind(context, deployable)

    @classmethod
    def get_suitable_ext_arq(cls, context, uuid):
        """From the inherit subclass find the suitable ExtARQ."""
        extarq = cls.get(context, uuid)
        typ = extarq.get_resources_from_device_profile_group()
        factory = cls.factory(typ)
        if factory != cls:
            return factory.get(context, uuid)
        return extarq

    def start_bind_job(self, context, valid_fields):
        """Check and start bind jobs for ARQ."""
        expected = ARQ_STATES_TRANSFORM_MATRIX[constants.ARQ_BIND_STARTED]
        # Check whether ARQ can be bound.
        if self.arq.state not in expected:
            raise exception.ARQBadState(
                state=self.arq.state, uuid=self.arq.uuid, expected=expected)

        hostname = valid_fields[self.arq.uuid]['hostname']
        devrp_uuid = valid_fields[self.arq.uuid]['device_rp_uuid']
        instance_uuid = valid_fields[self.arq.uuid]['instance_uuid']
        project_id = valid_fields[self.arq.uuid].get('project_id')
        LOG.info('[arqs:objs] bind. hostname: %(hostname)s, '
                 'devrp_uuid: %(devrp_uuid)s, '
                 'instance_uuid: %(instance_uuid)s, '
                 'project_id: %(project_id)s',
                 {'hostname': hostname, 'devrp_uuid': devrp_uuid,
                  'instance_uuid': instance_uuid, 'project_id': project_id})

        self.arq.hostname = hostname
        self.arq.device_rp_uuid = devrp_uuid
        self.arq.instance_uuid = instance_uuid
        self.arq.project_id = project_id

        # If prog fails, we'll change this ARQ state changes get committed here
        self.update_check_state(context, constants.ARQ_BIND_STARTED)

        dep = objects.Deployable.get_by_device_rp_uuid(context, devrp_uuid)
        return self._bind_job(context, dep)

    @classmethod
    def master(cls, context, arq_binds):
        """Start a master thread to monitor job workers."""
        jobs = {}
        instant = {}
        arq_uuids = [ea.arq.uuid for ea in arq_binds.keys()]
        for arq_uuid, job in arq_binds.items():
            kv = {arq_uuid: job}
            jobs.update(kv) if job else instant.update(kv)
        if not jobs:
            LOG.info("All ARQ(%s) bind process are instant.", arq_uuids)
            cls.check_bindings_result(context, arq_binds.keys())
            return
        th_workers = utils.ThreadWorks()
        works_generator = th_workers.get_workers_result(
            jobs.values(), timeout=CONF.bind_timeout)
        # arq_binds, timeout=1)
        LOG.info("Check ARQ(%s) bind jobs status.", arq_uuids)
        th_workers.spawn_master(
            cls.job_monitor, context, works_generator, arq_binds.keys())

    @classmethod
    def get_arq_bind_statuses(cls, arq_list):
        """Return the list of UUID/state pairs for each ARQ.

        :param arq_list: List of ARQ objects.
        :returns: List of (arq_uuid, arq_bind_status) tuples.
        :raises: ARQBadState exception if any ARQ state is invalid.
        """
        # NOTE(Sundar): This should be called when all ARQs have
        # completed binding, successfully or not. One or more may be in
        # deleting state. So, arq_state must be 'Bound', 'BindFailed'
        # or 'Deleting'
        state_map = constants.ARQ_BIND_STATES_STATUS_MAP
        good_states = list(state_map.keys())
        arq_bind_statuses = []
        for arq in arq_list:
            if arq.state not in good_states:
                raise exception.ARQBadState(
                    state=arq.state, uuid=arq.uuid, expected=good_states)
            arq_bind_status = (arq.uuid, state_map[arq.state])
            arq_bind_statuses.append(arq_bind_status)
        return arq_bind_statuses

    @classmethod
    def check_bindings_result(cls, context, instance_extarqs):
        """Check the ARQ bind status result.

        :param context: Security context
        :param instance_extarqs: List of ExtARQs for one instance
            at the time the async bind job was launched. Some of
            them may have been deleted since then: see note below.
        """
        # Batch get or get one by one? Maybe delete a ARQ
        arq_uuids = [ea.arq.uuid for ea in instance_extarqs]
        arq_list = [extarq.arq for extarq in instance_extarqs]

        instance_extarqs = list(instance_extarqs)
        instance_uuid = instance_extarqs[0].arq.instance_uuid

        # NOTE(Sundar): While the async ARQ bind job is in progress, there
        # could conceivably be calls to delete one or more of those ARQs.
        # One way to handle that is by pending the deletions till binds
        # complete. But the current implementation does not do that; so
        # we check for deleted ARQs here.
        # TODO(all): Revisit the handling of concurrent deletions.
        #   In the current design, Nova gets notified only after all ARQs
        # for an instance have completed binding, so concurrent deletions
        # are not expected from Nova.
        #   The extarq list() below returns only the currently available
        # (not deleted) ARQs among the specified ones.
        extarqs = cls.list(context, arq_uuids)
        if len(extarqs) < len(arq_uuids):
            LOG.error("ARQs(%s) bind status sync error, status is %s. "
                      "For some ARQs %s are deleted.",
                      arq_uuids, constants.ARQ_BIND_STATUS_FAILED,
                      set(arq_uuids) - set([ea.arq.uuid for ea in extarqs]))
            cls.bind_notify(instance_uuid, cls.get_arq_bind_statuses(arq_list))
            return

        status = constants.ARQ_BIND_STATUS_FINISH
        for extarq in extarqs:
            state = extarq.arq.state
            uuid = extarq.arq.uuid
            if state in constants.ARQ_PRE_BIND:
                # OPEN ignore ARQ_OUFOF_BIND_FLOW?
                status = constants.ARQ_BIND_STATUS_FAILED
                LOG.error("ARQs(%s) bind has not finished, status is %s.",
                          uuid, status)
                break
            elif state in constants.ARQ_OUFOF_BIND_FLOW + [
                constants.ARQ_BIND_FAILED]:
                # OPEN ignore ARQ_OUFOF_BIND_FLOW?
                status = constants.ARQ_BIND_STATUS_FAILED
                LOG.error("ARQs(%s) bind status sync error, status is %s.",
                          uuid, status)
                break
            elif state == constants.ARQ_BOUND:
                LOG.info("ARQs(%s) bind status sync finish, status is %s.",
                         uuid, status)
        if status == constants.ARQ_BIND_STATUS_FINISH:
            LOG.info('All ARQs %s async bind jobs has finished.', arq_uuids)
        cls.bind_notify(instance_uuid, cls.get_arq_bind_statuses(arq_list))

    @classmethod
    @utils.wrap_job_tb("Error in ARQ bind async job_monitor. Reason: %s")
    def job_monitor(cls, context, works_generator, extarqs):
        """monitor every deployable bind jobs."""
        # result: f.result(), f.exception_info(), f._state
        msg = None
        arq_uuids = [ea.arq.uuid for ea in extarqs]
        LOG.info('Monitor master check ARQ %s async bind job.', arq_uuids)
        for _, exc, _, err in works_generator:
            msg = exc or err
            if msg:
                LOG.error(msg)
            # TODO(Shaohe) Rollback? Such as We have _update_placement,
            # should cancel it.
        LOG.debug('Async bind job: join completed. ARQs: %s', arq_uuids)
        if not arq_uuids:
            return
        cls.check_bindings_result(context, extarqs)

    @classmethod
    def bind_notify(cls, instance_uuid, arq_bind_statuses):
        """Notify the bind status to nova."""
        nova_api = nova_client.NovaAPI()
        nova_api.notify_binding(instance_uuid, arq_bind_statuses)

    def get_resources_from_device_profile_group(self):
        """parser device profile group."""
        group = self.device_profile_group
        # example: {"resources:CUSTOM_ACCELERATOR_FPGA": "1"}
        resources = [
            (k.lstrip(constants.RESOURCES_PREFIX), v) for k, v in group.items()
            if k.startswith(constants.RESOURCES_PREFIX)]
        if not resources:
            raise exception.InvalidParameterValue(
                'No resources in device_profile_group: %s' % group)
        res_type = resources[0][0]
        return res_type

    @classmethod
    def apply_patch(cls, context, patch_list, valid_fields):
        """Apply JSON patch. See api/controllers/v2/arqs.py."""
        arq_binds = {}
        for arq_uuid, patch in patch_list.items():
            extarq = cls.get_suitable_ext_arq(context, arq_uuid)
            if patch[0]['op'] == 'add':  # All ops are 'add'
                # arq_notify_list.append(arq_uuid)
                job = extarq.start_bind_job(context, valid_fields)
                arq_binds[extarq] = job
            else:
                extarq.unbind(context)
        if arq_binds:
            cls.master(context, arq_binds)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from cyborg.api.controllers.v2 import versions
from cyborg.tests.unit.api import base as api_base


SERVICE_TYPE = 'accelerator'
H_MIN_VER = 'openstack-api-minimum-version'
H_MAX_VER = 'openstack-api-maximum-version'
H_RESP_VER = 'openstack-api-version'
MIN_VER = versions.min_version_string()
MAX_VER = versions.max_version_string()


class TestMicroversions(api_base.BaseApiTest):

    controller_list_response = [
        'id', 'links', 'max_version', 'min_version', 'status']

    def setUp(self):
        super(TestMicroversions, self).setUp()

    def test_wrong_major_version(self):
        response = self.get_json(
            '/v2',
            headers={'OpenStack-API-Version': '10'},
            expect_errors=True, return_json=False)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(406, response.status_int)
        expected_error_msg = ('Invalid value for'
                              ' OpenStack-API-Version header')
        self.assertTrue(response.json['error_message'])
        self.assertIn(expected_error_msg, response.json['error_message'])

    def test_without_specified_microversion(self):
        """If the header OpenStack-API-Version is absent in user's request,
        the default microversion is MIN_VER.
        """
        response = self.get_json('/v2', return_json=False)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        self.assertEqual(response.headers[H_RESP_VER], MIN_VER)
        self.assertTrue(all(x in response.json.keys() for x in
                            self.controller_list_response))

    def test_new_client_new_api(self):
        response = self.get_json(
            '/v2',
            headers={'OpenStack-API-Version': '2.0'},
            return_json=False)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        self.assertEqual(response.headers[H_RESP_VER], '2.0')
        self.assertTrue(all(x in response.json.keys() for x in
                            self.controller_list_response))

    def test_latest_microversion(self):
        response = self.get_json(
            '/v2',
            headers={'OpenStack-API-Version': 'latest'},
            return_json=False)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        self.assertEqual(response.headers[H_RESP_VER], MAX_VER)
        self.assertTrue(all(x in response.json.keys() for x in
                            self.controller_list_response))

    def test_unsupported_version(self):
        unsupported_version = str(float(MAX_VER) + 0.1)
        response = self.get_json(
            '/v2',
            headers={'OpenStack-API-Version': unsupported_version},
            expect_errors=True)
        self.assertEqual(406, response.status_int)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        expected_error_msg = ('Version %s was requested but the minor '
                              'version is not supported by this service. '
                              'The supported version range is' %
                              unsupported_version)
        self.assertTrue(response.json['error_message'])
        self.assertIn(expected_error_msg, response.json['error_message'])



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import os
import tempfile
import yaml

from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_upgradecheck import upgradecheck

from cyborg.cmd import status
from cyborg.common import authorize_wsgi
from cyborg.tests import base


class TestUpgradeCheckPolicyJSON(base.TestCase):

    def setUp(self):
        super(TestUpgradeCheckPolicyJSON, self).setUp()
        self.cmd = status.UpgradeCommands()
        authorize_wsgi.CONF.clear_override('policy_file', group='oslo_policy')
        self.data = {
            'rule_admin': 'True',
            'rule_admin2': 'is_admin:True'
        }
        self.temp_dir = self.useFixture(fixtures.TempDir())
        fd, self.json_file = tempfile.mkstemp(dir=self.temp_dir.path)
        fd, self.yaml_file = tempfile.mkstemp(dir=self.temp_dir.path)

        with open(self.json_file, 'w') as fh:
            jsonutils.dump(self.data, fh)
        with open(self.yaml_file, 'w') as fh:
            yaml.dump(self.data, fh)

        original_search_dirs = cfg._search_dirs

        def fake_search_dirs(dirs, name):
            dirs.append(self.temp_dir.path)
            return original_search_dirs(dirs, name)

        self.mock_search = self.useFixture(fixtures.MockPatch(
            'oslo_config.cfg._search_dirs')).mock
        self.mock_search.side_effect = fake_search_dirs

    def test_policy_json_file_fail_upgrade(self):
        # Test with policy json file full path set in config.
        self.flags(policy_file=self.json_file, group="oslo_policy")
        self.assertEqual(upgradecheck.Code.FAILURE,
                         self.cmd._check_policy_json().code)

    def test_policy_yaml_file_pass_upgrade(self):
        # Test with full policy yaml file path set in config.
        self.flags(policy_file=self.yaml_file, group="oslo_policy")
        self.assertEqual(upgradecheck.Code.SUCCESS,
                         self.cmd._check_policy_json().code)

    def test_no_policy_file_pass_upgrade(self):
        # Test with no policy file exist.
        self.assertEqual(upgradecheck.Code.SUCCESS,
                         self.cmd._check_policy_json().code)

    def test_default_policy_yaml_file_pass_upgrade(self):
        tmpfilename = os.path.join(self.temp_dir.path, 'policy.yaml')
        with open(tmpfilename, 'w') as fh:
            yaml.dump(self.data, fh)
        self.assertEqual(upgradecheck.Code.SUCCESS,
                         self.cmd._check_policy_json().code)

    def test_old_default_policy_json_file_fail_upgrade(self):
        self.flags(policy_file='policy.json', group="oslo_policy")
        tmpfilename = os.path.join(self.temp_dir.path, 'policy.json')
        with open(tmpfilename, 'w') as fh:
            jsonutils.dump(self.data, fh)
        self.assertEqual(upgradecheck.Code.FAILURE,
                         self.cmd._check_policy_json().code)



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils.fixture import uuidsentinel as uuids

from cyborg import context as cyborg_context
from cyborg.tests.unit.api.controllers.v2 import base as v2_test
from cyborg.tests.unit import policy_fixture


LOG = logging.getLogger(__name__)


class BasePolicyTest(v2_test.APITestV2):

    def setUp(self):
        super(BasePolicyTest, self).setUp()
        self.policy = self.useFixture(policy_fixture.PolicyFixture())

        self.admin_project_id = uuids.admin_project_id
        self.project_id = uuids.project_id
        self.foo_project_id = uuids.foo_project_id
        self.project_id_other = uuids.project_id_other

        # legacy default role: "default:admin_or_owner"
        self.legacy_admin_context = cyborg_context.RequestContext(
            user_id="legacy_admin", project_id=self.admin_project_id,
            roles='admin')
        self.legacy_owner_context = cyborg_context.RequestContext(
            user_id="legacy_owner", project_id=self.admin_project_id,
            roles='member')

        # system scoped users
        self.system_admin_context = cyborg_context.RequestContext(
            user_id="sys_admin",
            roles='admin', system_scope='all')

        self.system_member_context = cyborg_context.RequestContext(
            user_id="sys_member",
            roles='member', system_scope='all')

        self.system_reader_context = cyborg_context.RequestContext(
            user_id="sys_reader", roles='reader', system_scope='all')

        self.system_foo_context = cyborg_context.RequestContext(
            user_id="sys_foo", roles='foo', system_scope='all')

        # project scoped users
        self.project_admin_context = cyborg_context.RequestContext(
            user_id="project_admin", project_id=self.project_id,
            roles='admin')

        self.project_member_context = cyborg_context.RequestContext(
            user_id="project_member", project_id=self.project_id,
            roles='member')

        self.project_reader_context = cyborg_context.RequestContext(
            user_id="project_reader", project_id=self.project_id,
            roles='reader')

        self.project_foo_context = cyborg_context.RequestContext(
            user_id="project_foo", project_id=self.project_id,
            roles='foo')

        self.other_project_member_context = cyborg_context.RequestContext(
            user_id="other_project_member",
            project_id=self.project_id_other,
            roles='member')

        self.all_contexts = [
            self.legacy_admin_context, self.legacy_owner_context,
            self.system_admin_context, self.system_member_context,
            self.system_reader_context, self.system_foo_context,
            self.project_admin_context, self.project_member_context,
            self.project_reader_context, self.other_project_member_context,
            self.project_foo_context,
        ]



# Copyright 2020 Inspur, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Inspur FPGA driver implementation.
"""

import re

from oslo_concurrency import processutils
from oslo_serialization import jsonutils

from cyborg.accelerator.common import utils
from cyborg.common import constants
from cyborg.conf import CONF
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device
import cyborg.privsep

INSPUR_FPGA_FLAGS = ["Inspur Electronic Information Industry Co., Ltd.",
                     "Processing accelerators"]
INSPUR_FPGA_INFO_PATTERN = re.compile(
    r"(?P<devices>[0-9a-fA-F]{4}:[0-9a-fA-F]{2}:"
    r"[0-9a-fA-F]{2}\.[0-9a-fA-F]) "
    r"(?P<controller>.*) [\[].*]: (?P<model>.*) .*"
    r"[\[](?P<vendor_id>[0-9a-fA-F]"
    r"{4}):(?P<product_id>[0-9a-fA-F]{4})].*")

VENDOR_ID = "1bd4"
VENDOR_MAPS = {"1bd4": "inspur"}


@cyborg.privsep.sys_admin_pctxt.entrypoint
def lspci_privileged():
    cmd = ['lspci', '-nnn', '-D']
    return processutils.execute(*cmd)


def get_pci_devices(pci_flags, vendor_id=None):
    device_for_vendor_out = []
    all_device_out = []
    lspci_out = lspci_privileged()[0].split('\n')
    for i in range(len(lspci_out)):
        if any(x in lspci_out[i] for x in pci_flags):
            all_device_out.append(lspci_out[i])
            if vendor_id and vendor_id in lspci_out[i]:
                device_for_vendor_out.append(lspci_out[i])
    return device_for_vendor_out if vendor_id else all_device_out


def get_traits(vendor_id, product_id):
    """Generate traits for FPGAs.
    : param vendor_id: vendor_id of FPGA, eg."1bd4"
    : param product_id: product_id of FPGA, eg."a115".
    Example FPGA traits:
    {traits:["CUSTOM_FPGA_INSPUR", "CUSTOM_FPGA_PRODUCT_ID_A115"]}
    """
    traits = []
    traits.append("CUSTOM_FPGA_" + VENDOR_MAPS.get(vendor_id, "").upper())
    traits.append("CUSTOM_FPGA_PRODUCT_ID_" + product_id.upper())
    # TODO(wenping) Currently we don't support program by Cyborg. The operator
    # can bind Inspur FPGA card to guest and then use it in guest. So the BSP
    # data reported is ignored here now.
    return {"traits": traits}


def fpga_tree():
    fpga_list = []
    fpgas = get_pci_devices(INSPUR_FPGA_FLAGS, vendor_id=VENDOR_ID)
    for fpga in fpgas:
        m = INSPUR_FPGA_INFO_PATTERN.match(fpga)
        if m:
            fpga_dict = m.groupdict()
            # generate traits info
            traits = get_traits(
                fpga_dict["vendor_id"], fpga_dict["product_id"])
            fpga_dict["rc"] = constants.RESOURCES["FPGA"]
            fpga_dict.update(traits)
            fpga_list.append(_generate_driver_device(fpga_dict))
    return fpga_list


def _generate_driver_device(fpga):
    driver_device_obj = driver_device.DriverDevice()
    driver_device_obj.vendor = fpga["vendor_id"]
    driver_device_obj.model = fpga.get('model', 'miss model info')
    std_board_info = {'product_id': fpga.get('product_id'),
                      'controller': fpga.get('controller')}
    vendor_board_info = {
        'vendor_info': fpga.get('vendor_info', 'fpga_vb_info')}
    driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
    driver_device_obj.vendor_board_info = jsonutils.dumps(vendor_board_info)
    driver_device_obj.type = constants.DEVICE_FPGA
    driver_device_obj.stub = fpga.get('stub', False)
    driver_device_obj.controlpath_id = _generate_controlpath_id(fpga)
    driver_device_obj.deployable_list = _generate_dep_list(fpga)
    return driver_device_obj


def _generate_controlpath_id(fpga):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(fpga["devices"])
    return driver_cpid


def _generate_dep_list(fpga):
    driver_dep = driver_deployable.DriverDeployable()
    driver_dep.attribute_list = _generate_attribute_list(fpga)
    driver_dep.attach_handle_list = []
    # NOTE(wenping) Now simply named as <Compute_hostname>_<Device_address>
    # once cyborg needs to support Inspur FPGA devices discovered from a
    # baremetal node, we might need to support more formats.
    driver_dep.name = CONF.host + '_' + fpga["devices"]
    driver_dep.driver_name = VENDOR_MAPS.get(fpga["vendor_id"], '').upper()
    driver_dep.num_accelerators = 1
    driver_dep.attach_handle_list = [_generate_attach_handle(fpga)]
    return [driver_dep]


def _generate_attribute_list(fpga):
    attr_list = []
    for k, v in fpga.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key, driver_attr.value = k, v
            attr_list.append(driver_attr)
        if k == "traits":
            for index, val in enumerate(v):
                driver_attr = driver_attribute.DriverAttribute(
                    key="trait" + str(index), value=val)
                attr_list.append(driver_attr)
    return attr_list


def _generate_attach_handle(fpga):
    driver_ah = driver_attach_handle.DriverAttachHandle()
    driver_ah.attach_type = constants.AH_TYPE_PCI
    driver_ah.attach_info = utils.pci_str_to_json(fpga["devices"])
    driver_ah.in_use = False
    return driver_ah



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from cyborg.conf import agent
from cyborg.conf import api
from cyborg.conf import database
from cyborg.conf import default
from cyborg.conf import devices
from cyborg.conf import glance
from cyborg.conf import keystone
from cyborg.conf import nova
from cyborg.conf import placement
from cyborg.conf import service_token

CONF = cfg.CONF

api.register_opts(CONF)
agent.register_opts(CONF)
database.register_opts(CONF)
default.register_opts(CONF)
devices.register_opts(CONF)
service_token.register_opts(CONF)
glance.register_opts(CONF)
keystone.register_opts(CONF)
nova.register_opts(CONF)
placement.register_opts(CONF)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from cyborg.accelerator.drivers.aichip.huawei.ascend import AscendDriver
from cyborg.tests import base

d100_pci_res = (
    '0000:00:0c.0 Processing accelerators [1200]:'
    ' Device [19e5:d100] (rev 20)\n'
    '0000:00:0d.0 Processing accelerators [1200]:'
    ' Device [19e5:d100] (rev 20)\n',)


class TestAscendDriver(base.TestCase):

    @mock.patch('cyborg.accelerator.drivers.aichip.'
                'huawei.ascend.lspci_privileged',
                return_value=d100_pci_res)
    def test_discover(self, mock_pci):
        ascend_driver = AscendDriver()
        npu_list = ascend_driver.discover()
        self.assertEqual(2, len(npu_list))
        for ascend in npu_list:
            self.assertEqual('AICHIP', ascend.type)
            self.assertEqual('PCI', ascend.controlpath_id.cpid_type)
            self.assertEqual(json.loads(
                '{"class": "Processing accelerators", "device_id": "d100"}'),
                json.loads(ascend.std_board_info))
            self.assertEqual('19e5', ascend.vendor)

        self.assertEqual(
            {"device": "0c", "bus": "00", "domain": "0000", "function": "0"},
            json.loads(npu_list[0].controlpath_id.cpid_info))
        self.assertEqual(
            {"device": "0d", "bus": "00", "domain": "0000", "function": "0"},
            json.loads(npu_list[1].controlpath_id.cpid_info))

        self.assertEqual('Device_0000_00_0c_0',
                         npu_list[0].deployable_list[0].name)
        self.assertEqual('Device_0000_00_0d_0',
                         npu_list[1].deployable_list[0].name)



"""add_description_field_to_dps

Revision ID: 60d8ac91fd20
Revises: 7a4fd0fc3f8c
Create Date: 2020-01-19 16:15:04.231512

"""

# revision identifiers, used by Alembic.
revision = '60d8ac91fd20'
down_revision = '7a4fd0fc3f8c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('device_profiles', sa.Column('description',
                  sa.Text(), nullable=True))



# Copyright 2019 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import os
import re

from oslo_serialization import jsonutils

from cyborg.common import exception


_PCI_ADDRESS_PATTERN = ("^(hex{4}):(hex{2}):(hex{2}).(oct{1})$".
                        replace("hex", r"[\da-fA-F]").
                        replace("oct", "[0-7]"))
_PCI_ADDRESS_REGEX = re.compile(_PCI_ADDRESS_PATTERN)


def pci_str_to_json(pci_address, physnet=None):
    dbs, func = pci_address.split('.')
    domain, bus, slot = dbs.split(':')
    keys = ["domain", "bus", "device", "function"]
    values = [domain, bus, slot, func]
    if physnet:
        keys.append("physical_network")
        values.append(physnet)
    bdf_dict = dict(zip(keys, values))
    ordered_dict = collections.OrderedDict(sorted(bdf_dict.items()))
    bdf_json = jsonutils.dumps(ordered_dict)
    return bdf_json


def _get_sysfs_netdev_path(pci_addr, vf_interface=False):
    """Get the sysfs path based on the PCI address of the device.
    Assumes a networking device - will not check for the existence of the path.
    :param pci_addr: the pci addresee of the device(PF or VF).
    :param vf_interface: True if the pci_addr is a VF,
        False if the pci_addr is a PF.
    :returns: the sysfs path corresponds to the pci_addr.
    """
    if vf_interface:
        return "/sys/bus/pci/devices/%s/physfn/net" % pci_addr
    return "/sys/bus/pci/devices/%s/net" % pci_addr


def get_ifname_by_pci_address(pci_addr, vf_interface=False):
    """Get the interface name based on the pci address.
    The returned interface name is either the parent PF's or that of the PF
    itself based on the argument of vf_interface.
    :param pci_addr: the pci address of the device(PF or VF)
    :param vf_interface: True if the pci_addr is a VF,
        False if the pci_addr is a PF.
    :returns: the interface name corresponds to the pci_addr.
    """
    dev_path = _get_sysfs_netdev_path(pci_addr, vf_interface)
    try:
        dev_info = os.listdir(dev_path)
        return dev_info.pop()
    except Exception:
        raise exception.PciDeviceNotFoundById(id=pci_addr)


def parse_mappings(mapping_list):
    """Parse mapping devices list

    parses mapping device list in the form:
    physnet:pci_dev_1,pci_dev_2 or
    function:pci_dev_1,pci_dev_2
    :param mapping_list: list of string pairs in "key:value" format
                    the key part represents the physnet or function name
                    the value part is a list of device name separated by ","
    :returns: a dict of valid fields.
    """
    mapping = {}
    for dev_mapping in mapping_list:
        try:
            physnet_or_function, devices = dev_mapping.split(":", 1)
        except ValueError:
            raise ValueError(("Invalid mapping: '%s'") % dev_mapping)
        physnet_or_function = physnet_or_function.strip()
        if not physnet_or_function:
            raise ValueError(("Missing key in mapping: '%s'") % dev_mapping)
        if physnet_or_function in mapping:
            raise ValueError(
                ("Key %(physnet_or_function)s in mapping: %(mapping)s "
                 "not unique") % {'physnet_or_function': physnet_or_function,
                                  'mapping': dev_mapping})
        mapping[physnet_or_function] = set(dev.strip() for dev in
                                           devices.split("|") if dev.strip())
    return mapping


def get_vendor_maps():
    """The data is based on http://pci-ids.ucw.cz/read/PC/

    :return: vendor maps dict
    """
    return {"10de": "nvidia",
            "102b": "matrox",
            "1bd4": "inspur",
            "8086": "intel",
            "1099": "samsung",
            "1cf2": "zte"
            }


def mdev_str_to_json(pci_address, asked_type, vgpu_mark):
    dbs, func = pci_address.split('.')
    domain, bus, slot = dbs.split(':')
    keys = ["domain", "bus", "device", "function", "asked_type", "vgpu_mark"]
    values = [domain, bus, slot, func, asked_type, vgpu_mark]
    bdf_dict = dict(zip(keys, values))
    ordered_dict = collections.OrderedDict(sorted(bdf_dict.items()))
    bdf_json = jsonutils.dumps(ordered_dict)
    return bdf_json


def parse_address(address):
    """Returns (domain, bus, slot, function) from PCI address that is set
    in configuration
    """
    m = _PCI_ADDRESS_REGEX.match(address)
    if not m:
        raise exception.PciDeviceWrongAddressFormat(address=address)
    return m.groups()



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating AttachHandle via the DB API"""

from oslo_utils import uuidutils

from cyborg.common import exception
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils


class TestDbAttachHandle(base.DbTestCase):

    def test_create(self):
        random_uuid = uuidutils.generate_uuid()
        kw = {'uuid': random_uuid}
        created_ah = utils.create_test_attach_handle(self.context, **kw)
        self.assertEqual(random_uuid, created_ah['uuid'])

    def test_get_by_uuid(self):
        created_ah = utils.create_test_attach_handle(self.context)
        queried_ah = self.dbapi.attach_handle_get_by_uuid(
            self.context, created_ah['uuid'])
        self.assertEqual(created_ah['uuid'], queried_ah['uuid'])

    def test_get_by_id(self):
        created_ah = utils.create_test_attach_handle(self.context)
        queried_ah = self.dbapi.attach_handle_get_by_id(
            self.context, created_ah['id'])
        self.assertEqual(created_ah['id'], queried_ah['id'])

    def test_update(self):
        created_ah = utils.create_test_attach_handle(self.context)
        queried_ah = self.dbapi.attach_handle_update(
            self.context, created_ah['uuid'], {'attach_type': 'TEST_PCI'})
        self.assertEqual('TEST_PCI', queried_ah['attach_type'])

    def test_list(self):
        uuids = []
        for i in range(1, 4):
            ah = utils.create_test_attach_handle(
                self.context,
                id=i,
                uuid=uuidutils.generate_uuid())
            uuids.append(ah['uuid'])
        ahs = self.dbapi.attach_handle_list(self.context)
        ah_uuids = [item.uuid for item in ahs]
        self.assertEqual(sorted(uuids), sorted(ah_uuids))

    def test_list_by_type(self):
        ah1 = utils.create_test_attach_handle(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid(),
            attach_type='PCI')
        utils.create_test_attach_handle(
            self.context,
            id=2,
            uuid=uuidutils.generate_uuid(),
            attach_type='TEST_PCI')
        res = self.dbapi.attach_handle_list_by_type(
            self.context, attach_type='PCI')
        self.assertEqual(1, len(res))
        self.assertEqual(ah1['uuid'], res[0]['uuid'])

    def test_get_by_filters(self):
        ah1 = utils.create_test_attach_handle(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid(),
            deployable_id=1)
        utils.create_test_attach_handle(
            self.context,
            id=2,
            uuid=uuidutils.generate_uuid(),
            deployable_id=2)
        res = self.dbapi.attach_handle_get_by_filters(
            self.context, filters={"deployable_id": 1})
        self.assertEqual(1, len(res))
        self.assertEqual(ah1['uuid'], res[0]['uuid'])

    def test_allocate(self):
        utils.create_test_attach_handle(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid(),
            deployable_id=1)
        allocate_ah = self.dbapi.attach_handle_allocate(
            self.context, deployable_id=1)
        self.assertTrue(allocate_ah['in_use'])

    def test_delete(self):
        created_ah = utils.create_test_attach_handle(self.context)
        return_value = self.dbapi.attach_handle_delete(
            self.context,
            created_ah['uuid'])
        self.assertIsNone(return_value)

    def test_list_filter_is_none(self):
        """The main test is filters=None. If filters=None,
        it will be initialized to {}, that will return all attach
        handle same as the List Attach Handle API response.
        """
        ah1 = utils.create_test_attach_handle(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid())
        res = self.dbapi.attach_handle_get_by_filters(
            self.context, filters=None)
        self.assertEqual(1, len(res))
        self.assertEqual(ah1['uuid'], res[0]['uuid'])

    def test_get_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.attach_handle_get_by_uuid,
                          self.context, random_uuid)

    def test_delete_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.attach_handle_delete,
                          self.context, random_uuid)

    def test_do_allocate_attach_handle(self):
        dep_id = 100
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi._do_allocate_attach_handle,
                          self.context, dep_id)



# Copyright (c) 2018 Intel.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

"""Cyborg agent resource_tracker test cases."""
from cyborg.agent.resource_tracker import ResourceTracker
from cyborg.common import exception
from cyborg.conductor import rpcapi as cond_api
from cyborg.conf import CONF
from cyborg.tests import base


class TestResourceTracker(base.TestCase):
    """Test Agent ResourceTracker """

    def setUp(self):
        super(TestResourceTracker, self).setUp()
        self.host = CONF.host
        self.cond_api = cond_api.ConductorAPI()
        self.rt = ResourceTracker(self.host, self.cond_api)

    def test_update_usage(self):
        """Update the resource usage and stats after a change in an
        instance
        """
        # FIXME(Shaohe Feng) need add testcase. How to check the fpgas
        # has stored into DB by conductor correctly?
        pass

    def test_initialize_acc_drivers(self):
        enabled_drivers = ['intel_fpga_driver']
        self.rt._initialize_drivers(enabled_drivers=enabled_drivers)
        drivers = self.rt.acc_drivers
        self.assertEqual(len(drivers), len(enabled_drivers))

    def test_initialize_invalid_driver(self):
        enabled_drivers = ['invalid_driver']
        self.assertRaises(exception.InvalidDriver, self.rt._initialize_drivers,
                          enabled_drivers)

    @mock.patch('cyborg.agent.resource_tracker.LOG')
    def test_update_usage_failed_parent_provider(self, mock_log):
        with mock.patch.object(self.rt.conductor_api, 'report_data') as m:
            m.side_effect = exception.PlacementResourceProviderNotFound(
                resource_provider='foo')
            self.rt.update_usage(None)
            m.assert_called_once_with(None, 'fake-mini', mock.ANY)
        mock_log.error.assert_called_once_with('Unable to report usage: %s',
                                               m.side_effect)



"""placeholder

Revision ID: 22fb1af2d51e
Revises: 57539722e5cf
Create Date: 2019-12-03 00:56:06.023042

"""

# revision identifiers, used by Alembic.
revision = '22fb1af2d51e'
down_revision = '57539722e5cf'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



# Copyright 2020 Inspur Electronic Information Industry Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Utils for SSD driver.
"""

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_serialization import jsonutils

import re

from cyborg.accelerator.common import utils
from cyborg.common import constants
from cyborg.conf import CONF
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device
import cyborg.privsep

LOG = logging.getLogger(__name__)

SSD_FLAGS = ["Non-Volatile memory controller"]
SSD_INFO_PATTERN = re.compile(r"(?P<devices>[0-9a-fA-F]{4}:[0-9a-fA-F]{2}:"
                              r"[0-9a-fA-F]{2}\.[0-9a-fA-F]) "
                              r"(?P<controller>.*) [\[].*]: (?P<model>.*) .*"
                              r"[\[](?P<vendor_id>[0-9a-fA-F]"
                              r"{4}):(?P<product_id>[0-9a-fA-F]{4})].*")

VENDOR_MAPS = utils.get_vendor_maps()


@cyborg.privsep.sys_admin_pctxt.entrypoint
def lspci_privileged():
    cmd = ['lspci', '-nn', '-D']
    return processutils.execute(*cmd)


def get_pci_devices(pci_flags, vendor_id=None):
    device_for_vendor_out = []
    all_device_out = []
    lspci_out = lspci_privileged()[0].split('\n')
    for i in range(len(lspci_out)):
        if any(x in lspci_out[i] for x in pci_flags):
            all_device_out.append(lspci_out[i])
            if vendor_id and vendor_id in lspci_out[i]:
                device_for_vendor_out.append(lspci_out[i])
    return device_for_vendor_out if vendor_id else all_device_out


def get_traits(vendor_id, product_id):
    """Generate traits for SSDs.
    : param vendor_id: vendor_id of SSD, eg."1bd4"
    : param product_id: product_id of SSD, eg."1003".
    Example SSD traits:
    {traits:["CUSTOM_SSD_INSPUR", "CUSTOM_SSD_PRODUCT_ID_1003"]}
    """
    traits = []
    traits.append("CUSTOM_SSD_" + VENDOR_MAPS.get(vendor_id, "").upper())
    traits.append("CUSTOM_SSD_PRODUCT_ID_" + product_id.upper())
    return {"traits": traits}


def discover_vendors():
    vendors = set()
    ssds = get_pci_devices(SSD_FLAGS)
    for ssd in ssds:
        m = SSD_INFO_PATTERN.match(ssd)
        if m:
            vendor_id = m.groupdict().get("vendor_id")
            vendors.add(vendor_id)
    return vendors


def discover_ssds(vendor_id=None):
    ssd_list = []
    ssds = get_pci_devices(SSD_FLAGS, vendor_id)
    for ssd in ssds:
        m = SSD_INFO_PATTERN.match(ssd)
        if m:
            ssd_dict = m.groupdict()
            ssd_dict['hostname'] = CONF.host
            # generate traits info
            traits = get_traits(ssd_dict["vendor_id"], ssd_dict["product_id"])
            ssd_dict["rc"] = constants.RESOURCES["SSD"]
            ssd_dict.update(traits)
            ssd_list.append(_generate_driver_device(ssd_dict))
    return ssd_list


def _generate_driver_device(ssd):
    driver_device_obj = driver_device.DriverDevice()
    driver_device_obj.vendor = ssd["vendor_id"]
    driver_device_obj.model = ssd.get('model', 'miss model info')
    std_board_info = {'product_id': ssd.get('product_id'),
                      'controller': ssd.get('controller')}
    vendor_board_info = {'vendor_info': ssd.get('vendor_info', 'ssd_vb_info')}
    driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
    driver_device_obj.vendor_board_info = jsonutils.dumps(vendor_board_info)
    driver_device_obj.type = constants.DEVICE_SSD
    driver_device_obj.stub = ssd.get('stub', False)
    driver_device_obj.controlpath_id = _generate_controlpath_id(ssd)
    driver_device_obj.deployable_list = _generate_dep_list(ssd)
    return driver_device_obj


def _generate_controlpath_id(ssd):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    # NOTE: SSDs , they all report "PCI" as
    # their cpid_type, while attach_handle_type of them are different.
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(ssd["devices"])
    return driver_cpid


def _generate_dep_list(ssd):
    driver_dep = driver_deployable.DriverDeployable()
    driver_dep.attribute_list = _generate_attribute_list(ssd)
    driver_dep.attach_handle_list = []
    # NOTE(wenping) Now simply named as <Compute_hostname>_<Device_address>
    # once cyborg needs to support SSD devices discovered from a baremetal
    # node, we might need to support more formats.
    driver_dep.name = ssd.get('hostname', '') + '_' + ssd["devices"]
    driver_dep.driver_name = VENDOR_MAPS.get(ssd["vendor_id"], '').upper()
    # driver_dep.num_accelerators for SSD is 1
    driver_dep.num_accelerators = 1
    driver_dep.attach_handle_list = [_generate_attach_handle(ssd)]
    return [driver_dep]


def _generate_attach_handle(ssd):
    driver_ah = driver_attach_handle.DriverAttachHandle()
    if ssd["rc"] == "CUSTOM_SSD":
        driver_ah.attach_type = constants.AH_TYPE_PCI
    else:
        driver_ah.attach_type = constants.AH_TYPE_MDEV
    driver_ah.in_use = False
    driver_ah.attach_info = utils.pci_str_to_json(ssd["devices"])
    return driver_ah


def _generate_attribute_list(ssd):
    attr_list = []
    for k, v in ssd.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key, driver_attr.value = k, v
            attr_list.append(driver_attr)
        if k == "traits":
            values = ssd.get(k, [])
            for index, val in enumerate(values):
                driver_attr = driver_attribute.DriverAttribute(
                    key="trait" + str(index), value=val)
                attr_list.append(driver_attr)
    return attr_list



# Copyright 2018 Lenovo (Beijing) Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import base as object_base

from cyborg.objects import base
from cyborg.objects.control_path import ControlpathID
from cyborg.objects.device import Device
from cyborg.objects.driver_objects.driver_controlpath_id import \
    DriverControlPathID
from cyborg.objects.driver_objects.driver_deployable import DriverDeployable
from cyborg.objects import fields as object_fields


@base.CyborgObjectRegistry.register
class DriverDevice(base.DriverObjectBase,
                   object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vendor': object_fields.StringField(nullable=False),
        'model': object_fields.StringField(nullable=False),
        'type': object_fields.DeviceTypeField(nullable=False),
        'std_board_info': object_fields.StringField(nullable=True),
        # vendor board info should be a dict for driver-specific resource
        # provider.
        'vendor_board_info': object_fields.StringField(nullable=True),
        # hostname will be set by the agent, so driver don't need to report.
        # Each controlpath_id corresponds to a different PF. For now
        # we are sticking with a single cpid.
        'controlpath_id': object_fields.ObjectField('DriverControlPathID',
                                                    nullable=False),
        'deployable_list': object_fields.ListOfObjectsField('DriverDeployable',
                                                            default=[],
                                                            nullable=False),
        'stub': object_fields.BooleanField(nullable=False, default=False)
    }

    def create(self, context, host):
        """Create a driver-side Device Object into DB. This object will be
        stored in many db tables: device, deployable, attach_handle,
        controlpath_id etc. by calling related Object.
        """
        # first store in device table through Device Object.

        device_obj = Device(context=context,
                            type=self.type,
                            vendor=self.vendor,
                            model=self.model,
                            hostname=host
                            )
        if hasattr(self, 'std_board_info'):
            device_obj.std_board_info = self.std_board_info
        if hasattr(self, 'vendor_board_info'):
            device_obj.vendor_board_info = self.vendor_board_info
        device_obj.create(context)

        # for the controlpath_id, call driver_controlpath_id to create.
        cpid_obj = self.controlpath_id.create(context, device_obj.id)
        # for deployable_list, call internal layer object: driver_deployable
        # to create.
        for driver_deployable in self.deployable_list:
            driver_deployable.create(context, device_obj.id, cpid_obj.id)

    def destroy(self, context, host):
        """Delete a driver-side Device Object from db. This should
        delete the internal layer objects.
        """
        # get dev_obj_list from hostname
        device_obj = self.get_device_obj(context, host)
        # delete deployable_list first.
        for driver_deployable in self.deployable_list:
            driver_deployable.destroy(context, device_obj.id)
        if hasattr(self.controlpath_id, 'cpid_info'):
            cpid_obj = ControlpathID.get_by_device_id_cpidinfo(
                context, device_obj.id, self.controlpath_id.cpid_info)
            # delete controlpath_id
            cpid_obj.destroy(context)
        # delete the device
        device_obj.destroy(context)

    def get_device_obj(self, context, host):
        """Get a driver-side Device Object from db.

        :param context: requested context.
        :param host: hostname of the node.
        :return: a device object of current driver device object. It will
        return on value because it has controlpath_id.
        """
        # get dev_obj_list from hostname
        device_obj_list = Device.get_list_by_hostname(context, host)
        # use controlpath_id.cpid_info to identiy one Device.
        for device_obj in device_obj_list:
            # get cpid_obj, could be empty or only one value.
            cpid_obj = ControlpathID.get_by_device_id_cpidinfo(
                context, device_obj.id, self.controlpath_id.cpid_info)
            # find the one cpid_obj with cpid_info
            if cpid_obj is not None:
                return device_obj

    @classmethod
    def list(cls, context, host):
        """Form driver-side device object list from DB for one host.
        A list may contains driver_device_object without controlpath_id.(In
        the case some of controlpath_id can't store successfully but its
        devices stores successfully.)
        """
        # get dev_obj_list from hostname
        dev_obj_list = Device.get_list_by_hostname(context, host)
        driver_dev_obj_list = []
        for dev_obj in dev_obj_list:
            cpid = DriverControlPathID.get(context, dev_obj.id)
            # NOTE: will not return device without controlpath_id.
            if cpid is not None:
                driver_dev_obj = \
                    cls(context=context, vendor=dev_obj.vendor,
                        model=dev_obj.model, type=dev_obj.type,
                        std_board_info=dev_obj.std_board_info,
                        vendor_board_info=dev_obj.vendor_board_info,
                        controlpath_id=cpid,
                        deployable_list=DriverDeployable.list(context,
                                                              dev_obj.id)
                        )
                driver_dev_obj_list.append(driver_dev_obj)
        return driver_dev_obj_list

    def get_device_obj_by_device_id(self, context, device_id):
        """Get device object by device id.

        :param context: requested context.
        :param host: hostname of the node.
        :return: a device object of current driver device object. It will
        return on value because it has controlpath_id.
        """
        # get dev_obj_list from hostname
        device_obj = Device.get_by_device_id(context, device_id)
        # use controlpath_id.cpid_info to identiy one Device.
        # get cpid_obj, could be empty or only one value.
        ControlpathID.get_by_device_id_cpidinfo(
            context, device_obj.id, self.controlpath_id.cpid_info)
        # find the one cpid_obj with cpid_info
        return device_obj



# Copyright 2019 Intel, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg import objects
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils
from cyborg.tests.unit import fake_device_profile


class TestDeviceProfileObject(base.DbTestCase):

    def setUp(self):
        super(TestDeviceProfileObject, self).setUp()
        self.fake_device_profile = utils.get_test_device_profile()

    def test_get_by_name(self):
        name = self.fake_device_profile['name']
        with mock.patch.object(self.dbapi, 'device_profile_get',
                               autospec=True) as mock_db_devprof_get:
            mock_db_devprof_get.return_value = self.fake_device_profile
            obj_devprof = objects.DeviceProfile.get_by_name(self.context, name)
            mock_db_devprof_get.assert_called_once_with(self.context, name)
            self.assertEqual(self.context, obj_devprof._context)
            self.assertEqual(name, obj_devprof.name)
            self.assertIn('description', obj_devprof)

    def test_get_by_uuid(self):
        uuid = self.fake_device_profile['uuid']
        with mock.patch.object(self.dbapi, 'device_profile_get_by_uuid',
                               autospec=True) as mock_db_devprof_get:
            mock_db_devprof_get.return_value = self.fake_device_profile
            obj_devprof = objects.DeviceProfile.get_by_uuid(self.context, uuid)
            mock_db_devprof_get.assert_called_once_with(self.context, uuid)
            self.assertEqual(self.context, obj_devprof._context)
            self.assertEqual(uuid, obj_devprof.uuid)
            self.assertIn('description', obj_devprof)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'device_profile_list',
                               autospec=True) as mock_db_devprof_list:
            mock_db_devprof_list.return_value = [self.fake_device_profile]
            obj_devprofs = objects.DeviceProfile.list(self.context)
            self.assertEqual(1, mock_db_devprof_list.call_count)
            self.assertEqual(1, len(obj_devprofs))
            self.assertIsInstance(obj_devprofs[0], objects.DeviceProfile)
            self.assertEqual(self.context, obj_devprofs[0]._context)
            self.assertEqual(self.fake_device_profile['name'],
                             obj_devprofs[0].name)
            self.assertEqual(self.fake_device_profile['description'],
                             obj_devprofs[0].description)

    def test_create(self):
        api_devprofs = fake_device_profile.get_api_devprofs()
        api_devprof = api_devprofs[0]
        db_devprofs = fake_device_profile.get_db_devprofs()
        db_devprof = db_devprofs[0]
        with mock.patch.object(self.dbapi, 'device_profile_create',
                               autospec=True) as mock_db_devprof_create:
            mock_db_devprof_create.return_value = self.fake_device_profile
            obj_devprof = objects.DeviceProfile(**api_devprof)
            obj_devprof.create(self.context)
            mock_db_devprof_create.assert_called_once_with(
                self.context, db_devprof)

    def test_destroy(self):
        uuid = self.fake_device_profile['uuid']
        with mock.patch.object(self.dbapi, 'device_profile_get_by_uuid',
                               autospec=True) as mock_dp_get:
            mock_dp_get.return_value = self.fake_device_profile
            with mock.patch.object(self.dbapi, 'device_profile_delete',
                                   autospec=True) as m_dp_delete:
                m_dp_delete.return_value = None
                obj_devprof = objects.DeviceProfile.get_by_uuid(
                    self.context, uuid)
                obj_devprof.destroy(self.context)
                m_dp_delete.assert_called_once_with(self.context, uuid)
                self.assertEqual(self.context, obj_devprof._context)

    def test_update(self):
        fake_db_devprofs = fake_device_profile.get_db_devprofs()
        fake_obj_devprofs = fake_device_profile.get_obj_devprofs()
        db_devprof = fake_db_devprofs[0]
        db_devprof['created_at'] = None
        db_devprof['updated_at'] = None
        with mock.patch.object(self.dbapi, 'device_profile_get_by_uuid',
                               autospec=True) as mock_dp_get:
            mock_dp_get.return_value = db_devprof
            uuid = fake_db_devprofs[0]['uuid']
            # Start with db_devprofs[0], corr. to fake_obj_devprofs[0]
            obj_devprof = objects.DeviceProfile.get_by_uuid(self.context, uuid)
            # Change contents to fake_obj_devprofs[1] except uuid
            obj_devprof = fake_obj_devprofs[1]
            obj_devprof['uuid'] = uuid
            with mock.patch.object(self.dbapi, 'device_profile_update',
                                   autospec=True) as mock_dp_update:
                mock_dp_update.return_value = db_devprof
                obj_devprof.save(self.context)
                mock_dp_get.assert_called_once_with(self.context, uuid)
                mock_dp_update.assert_called_once()

    def test_obj_make_compatible(self):
        dp_obj = objects.DeviceProfile(description="fake description")
        primitive = dp_obj.obj_to_primitive()
        dp_obj.obj_make_compatible(primitive['cyborg_object.data'], '1.0')
        self.assertNotIn('description', primitive['cyborg_object.data'])
        primitive = dp_obj.obj_to_primitive()
        dp_obj.obj_make_compatible(primitive['cyborg_object.data'], '1.1')
        self.assertIn('description', primitive['cyborg_object.data'])



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cyborg.common import exception
from cyborg.common import utils
from keystoneauth1 import exceptions as ks_exc
import os_resource_classes as orc
from oslo_log import log as logging
from oslo_middleware import request_id


LOG = logging.getLogger(__name__)
NESTED_PROVIDER_API_VERSION = '1.14'
POST_RPS_RETURNS_PAYLOAD_API_VERSION = '1.20'
PLACEMENT_CLIENT_SEMAPHORE = 'placement_client'


class PlacementClient(object):
    """Client class for reporting to placement."""

    def __init__(self):
        self._client = utils.get_sdk_adapter('placement')

    def get(self, url, version=None, global_request_id=None):
        res = self._client.get(url, microversion=version,
                               global_request_id=global_request_id)
        if res and res.status_code >= 500:
            raise exception.PlacementServerError(
                "Placement Server has some error at this time.")
        LOG.debug('Successfully get resources from placement: %s', url)
        return res

    def post(self, url, data, version=None, global_request_id=None):
        res = self._client.post(url, json=data, microversion=version,
                                global_request_id=global_request_id)
        if res.status_code >= 500:
            raise exception.PlacementServerError(
                "Placement Server has some error at this time.")
        LOG.debug('Successfully create resources from placement: %s', url)
        return res

    def put(self, url, data, version=None, global_request_id=None):
        kwargs = {}
        if data is not None:
            kwargs['json'] = data
        res = self._client.put(url, microversion=version,
                               global_request_id=global_request_id,
                               **kwargs)
        if res.status_code >= 500:
            raise exception.PlacementServerError(
                "Placement Server has some error at this time.")
        LOG.debug('Successfully update resources from placement: %s', url)
        return res

    def delete(self, url, version=None, global_request_id=None):
        res = self._client.delete(url, microversion=version,
                                  global_request_id=global_request_id)
        if res.status_code >= 500:
            raise exception.PlacementServerError(
                "Placement Server has some error at this time.")
        LOG.debug('Successfully delete resources from placement: %s', url)
        return res

    def _get_rp_traits(self, rp_uuid):
        resp = self.get("/resource_providers/%s/traits" % rp_uuid,
                        version='1.6')
        if resp.status_code != 200:
            raise Exception(
                "Failed to get traits for rp %s: HTTP %d: %s" %
                (rp_uuid, resp.status_code, resp.text))
        return resp.json()

    def _ensure_traits(self, trait_names):
        # TODO(Xinran): maintain a reference count of how many RPs use
        # this trait and do the deletion only when the last RP is deleted.
        for trait_name in trait_names:
            trait = self.get("/traits/%s" % trait_name, version='1.6')
            if trait:
                LOG.info("Trait %(trait)s already existed",
                         {"trait": trait_name})
                continue
            resp = self.put("/traits/%s" % trait_name, None, version='1.6')
            if resp.status_code == 201:
                LOG.info("Created trait %(trait)s", {"trait": trait_name})
            else:
                raise Exception(
                    "Failed to create trait %s: HTTP %d: %s" %
                    (trait_name, resp.status_code, resp.text))

    def _put_rp_traits(self, rp_uuid, traits_json):
        generation = self.get_resource_provider(
            resource_provider_uuid=rp_uuid)['generation']
        payload = {
            'resource_provider_generation': generation,
            'traits': traits_json["traits"],
        }
        resp = self.put(
            "/resource_providers/%s/traits" % rp_uuid, payload, version='1.6')
        if resp.status_code != 200:
            raise Exception(
                "Failed to set traits to %s for rp %s: HTTP %d: %s" %
                (traits_json, rp_uuid, resp.status_code, resp.text))

    def add_traits_to_rp(self, rp_uuid, trait_names):
        self._ensure_traits(trait_names)
        traits_json = self._get_rp_traits(rp_uuid)
        traits = list(set(traits_json['traits'] + trait_names))
        traits_json['traits'] = traits
        self._put_rp_traits(rp_uuid, traits_json)

    def delete_trait_by_name(self, context, rp_uuid, trait_name):
        traits_json = self._get_rp_traits(rp_uuid)
        traits = [
            trait for trait in traits_json['traits']
            if trait != trait_name
            ]
        traits_json['traits'] = traits
        self._put_rp_traits(rp_uuid, traits_json)
        self._delete_trait(context, trait_name)

    def delete_traits_with_prefixes(self, context, rp_uuid, trait_prefixes):
        traits_json = self._get_rp_traits(rp_uuid)
        traits = [
            trait for trait in traits_json['traits']
            if not any(trait.startswith(prefix)
                       for prefix in trait_prefixes)]
        delete_traits = set(traits_json['traits']) - set(traits)
        traits_json['traits'] = traits
        self._put_rp_traits(rp_uuid, traits_json)
        for trait in delete_traits:
            self._delete_trait(context, trait)

    def get_placement_request_id(self, response):
        if response is not None:
            return response.headers.get(request_id.HTTP_RESP_HEADER_REQUEST_ID)

    def update_inventory(
            self, resource_provider_uuid, inventories,
            resource_provider_generation=None, version=None):
        if resource_provider_generation is None:
            resource_provider_generation = self.get_resource_provider(
                resource_provider_uuid=resource_provider_uuid)['generation']
        url = '/resource_providers/%s/inventories' % resource_provider_uuid
        body = {
            'resource_provider_generation': resource_provider_generation,
            'inventories': inventories
        }
        try:
            return self.put(url, body, version=version).json()
        except ks_exc.NotFound:
            raise exception.PlacementResourceProviderNotFound(
                resource_provider=resource_provider_uuid)

    def get_resource_provider(self, resource_provider_uuid):
        """Get resource provider by UUID.

        :param resource_provider_uuid: UUID of the resource provider.
        :raises PlacementResourceProviderNotFound: For failure to find resource
        :returns: The Resource Provider matching the UUID.
        """
        url = '/resource_providers/%s' % resource_provider_uuid
        try:
            return self.get(url).json()
        except ks_exc.NotFound:
            raise exception.PlacementResourceProviderNotFound(
                resource_provider=resource_provider_uuid)

    def _create_resource_provider(self, context, uuid, name,
                                  parent_provider_uuid=None):
        """Calls the placement API to create a new resource provider record.

        :param context: The security context
        :param uuid: UUID of the new resource provider
        :param name: Name of the resource provider
        :param parent_provider_uuid: Optional UUID of the immediate parent
        :return: A dict of resource provider information object representing
                 the newly-created resource provider.
        :raise: ResourceProviderCreationFailed or
                ResourceProviderRetrievalFailed on error.
        """
        url = "/resource_providers"
        payload = {
            'uuid': uuid,
            'name': name,
        }
        if parent_provider_uuid is not None:
            payload['parent_provider_uuid'] = parent_provider_uuid

        # Bug #1746075: First try the microversion that returns the new
        # provider's payload.
        resp = self.post(url, payload,
                         version=POST_RPS_RETURNS_PAYLOAD_API_VERSION,
                         global_request_id=context.global_id)

        placement_req_id = self.get_placement_request_id(resp)

        if resp:
            msg = ("[%(placement_req_id)s] Created resource provider record "
                   "via placement API for resource provider with UUID "
                   "%(uuid)s and name %(name)s.")
            args = {
                'uuid': uuid,
                'name': name,
                'placement_req_id': placement_req_id,
            }
            LOG.info(msg, args)
            return resp.json()

    def ensure_resource_provider(self, context, uuid, name=None,
                                 parent_provider_uuid=None):
        resp = self.get("/resource_providers/%s" % uuid, version='1.6')
        if resp.status_code == 200:
            LOG.info("Resource Provider %(uuid)s already exists",
                     {"uuid": uuid})
        else:
            LOG.info("Creating resource provider %(provider)s",
                     {"provider": name or uuid})
            try:
                resp = self._create_resource_provider(context, uuid, name,
                                                      parent_provider_uuid)
            except Exception:
                raise exception.ResourceProviderCreationFailed(
                    name=name or uuid)
        return uuid

    def ensure_resource_classes(self, context, names):
        """Make sure resource classes exist."""
        version = '1.7'
        to_ensure = set(names)
        for name in to_ensure:
            # no payload on the put request
            # if rc exists in placement's db, skip it.
            if name in orc.STANDARDS:
                return
            resp = self.put(
                "/resource_classes/%s" % name, None, version=version,
                global_request_id=context.global_id)
            if not resp:
                msg = ("Failed to ensure resource class record with placement "
                       "API for resource class %(rc_name)s. Got "
                       "%(status_code)d: %(err_text)s.")
                args = {
                    'rc_name': name,
                    'status_code': resp.status_code,
                    'err_text': resp.text,
                }
                LOG.error(msg, args)
                raise exception.InvalidResourceClass(resource_class=name)
            elif resp.status_code == 204:
                LOG.info("Resource class  %(rc_name)s already exists",
                         {"rc_name": name})
            elif resp.status_code == 201:
                LOG.info("Successfully created resource class %(rc_name).", {
                         "rc_name", name})

    def get_providers_in_tree(self, context, uuid):
        """Queries the placement API for a list of the resource providers in
        the tree associated with the specified UUID.

        :param context: The security context
        :param uuid: UUID identifier for the resource provider to look up
        :return: A list of dicts of resource provider information, which may be
                 empty if no provider exists with the specified UUID.
        :raise: ResourceProviderRetrievalFailed on error.
        """
        resp = self.get("/resource_providers?in_tree=%s" % uuid,
                        version=NESTED_PROVIDER_API_VERSION,
                        global_request_id=context.global_id)

        if resp.status_code == 200:
            return resp.json()['resource_providers']

        # Some unexpected error
        placement_req_id = self.get_placement_request_id(resp)
        msg = ("[%(placement_req_id)s] Failed to retrieve resource provider "
               "tree from placement API for UUID %(uuid)s. Got "
               "%(status_code)d: %(err_text)s.")
        args = {
            'uuid': uuid,
            'status_code': resp.status_code,
            'err_text': resp.text,
            'placement_req_id': placement_req_id,
        }
        LOG.error(msg, args)
        raise exception.ResourceProviderRetrievalFailed(uuid=uuid)

    def delete_provider(self, rp_uuid, global_request_id=None):
        resp = self.delete('/resource_providers/%s' % rp_uuid,
                           global_request_id=global_request_id)
        # Check for 404 since we don't need to warn/raise if we tried to delete
        # something which doesn"t actually exist.
        if resp.ok:
            LOG.info("Deleted resource provider %s", rp_uuid)
            return

        msg = ("[%(placement_req_id)s] Failed to delete resource provider "
               "with UUID %(uuid)s from the placement API. Got "
               "%(status_code)d: %(err_text)s.")
        args = {
            'placement_req_id': self.get_placement_request_id(resp),
            'uuid': rp_uuid,
            'status_code': resp.status_code,
            'err_text': resp.text
        }
        LOG.error(msg, args)
        # On conflict, the caller may wish to delete allocations and
        # redrive.  (Note that this is not the same as a
        # PlacementAPIConflict case.)
        if resp.status_code == 409:
            raise exception.ResourceProviderInUse()
        raise exception.ResourceProviderDeletionFailed(uuid=rp_uuid)

    def delete_rc_by_name(self, context, name):
        """Delete resource class from placement by name."""
        resp = self.delete(
            "/resource_classes/%s" % name, global_request_id=context.global_id)
        if not resp:
            msg = ("Failed to delete resource class record with placement "
                   "API for resource class %(rc_name)s. Got "
                   "%(status_code)d: %(err_text)s.")
            args = {
                'rc_name': name,
                'status_code': resp.status_code,
                'err_text': resp.text,
            }
            LOG.error(msg, args)
        elif resp.status_code == 204:
            LOG.info("Successfully delete resource class %(rc_name).", {
                     "rc_name", name})

    def _delete_trait(self, context, name):
        """Delete trait from placement by name."""
        version = '1.6'
        resp = self.delete("/traits/%s" % name, version=version,
                           global_request_id=context.global_id)
        if not resp:
            msg = ("Failed to delete trait record with placement "
                   "API for trait %(trait_name)s. Got "
                   "%(status_code)d: %(err_text)s.")
            args = {
                'trait_name': name,
                'status_code': resp.status_code,
                'err_text': resp.text,
            }
            LOG.error(msg, args)
        elif resp.status_code == 204:
            LOG.info("Successfully delete trait %(trait_name).", {
                     "trait_name", name})

    def update_rp_inventory_reserved(self, rp_uuid, resource, total, reserved):
        update_inventory = {resource: {"total": total, "reserved": reserved}}
        self.update_inventory(rp_uuid, update_inventory, version='1.26')



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_upgradecheck import upgradecheck
from oslo_utils import fileutils

from cyborg.common.i18n import _


CONF = cfg.CONF


class Checks(upgradecheck.UpgradeCommands):

    """Various upgrade checks should be added as separate methods in this class
    and added to _upgrade_checks tuple.
    """

    def _check_policy_json(self):
        "Checks to see if policy file is JSON-formatted policy file."
        msg = _("Your policy file is JSON-formatted which is "
                "deprecated since Victoria release (Cyborg 5.0.0). "
                "You need to switch to YAML-formatted file. You can use the "
                "``oslopolicy-convert-json-to-yaml`` tool to convert existing "
                "JSON-formatted files to YAML-formatted files in a "
                "backwards-compatible manner: "
                "https://docs.openstack.org/oslo.policy/"
                "latest/cli/oslopolicy-convert-json-to-yaml.html.")
        status = upgradecheck.Result(upgradecheck.Code.SUCCESS)
        # NOTE(gmann): Check if policy file exist and is in
        # JSON format by actually loading the file not just
        # by checking the extension.
        policy_path = CONF.find_file(CONF.oslo_policy.policy_file)
        if policy_path and fileutils.is_json(policy_path):
            status = upgradecheck.Result(upgradecheck.Code.FAILURE, msg)
        return status

    # The format of the check functions is to return an
    # oslo_upgradecheck.upgradecheck.Result
    # object with the appropriate
    # oslo_upgradecheck.upgradecheck.Code and details set.
    # If the check hits warnings or failures then those should be stored
    # in the returned Result's "details" attribute. The
    # summary will be rolled up at the end of the check() method.
    _upgrade_checks = (
        # Added in Victoria
        (_('Policy File JSON to YAML Migration'), _check_policy_json),
    )


def main():
    return upgradecheck.main(
        cfg.CONF, project='cyborg', upgrade_command=Checks())


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2021 Inspur.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from cyborg import objects
from cyborg.objects import attach_handle
from cyborg.objects import fields


def get_fake_attach_handle_as_dict():
    attach_handle1 = {
        'id': 1,
        'uuid': uuidutils.generate_uuid(),
        'in_use': 0,
        'cp_id': 1,
        'deployable_id': 1,
        'attach_type': "PCI",
        'attach_info': '{"domain": "0000", "bus": "0c",'
                       '"device": "0", "function": "1"}',
        }

    attach_handle2 = {
        'id': 2,
        'uuid': uuidutils.generate_uuid(),
        'in_use': 1,
        'cp_id': 2,
        'deployable_id': 2,
        'attach_type': "PCI",
        'attach_info': '{"domain": "0000", "bus": "0c",'
                       '"device": "0", "function": "1"}',
        }

    return [attach_handle1, attach_handle2]


def _convert_from_dict_to_obj(ah_dict):
    obj_ah = attach_handle.AttachHandle()
    for field in ah_dict.keys():
        obj_ah[field] = ah_dict[field]
    return obj_ah


def _convert_to_db_ah(ah_dict):
    for name, field in objects.AttachHandle.fields.items():
        if name in ah_dict:
            continue
        if field.nullable:
            ah_dict[name] = None
        elif field.default != fields.UnspecifiedDefault:
            ah_dict[name] = field.default
        else:
            raise Exception('fake_db_attach_handle needs help with %s' % name)
    return ah_dict


def get_db_attach_handles():
    ahs_list = get_fake_attach_handle_as_dict()
    db_ahs = list(map(_convert_to_db_ah, ahs_list))
    return db_ahs


def get_fake_attach_handle_objs():
    ahs_list = get_fake_attach_handle_as_dict()
    obj_ahs = list(map(_convert_from_dict_to_obj, ahs_list))
    return obj_ahs
